# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
"""Parametrizable families of optimizers.

Caution
-------
This module and its available classes are experimental and may change quickly in the near future.
"""
from .optimizerlib import ParametrizedOnePlusOne
from .optimizerlib import ParametrizedTBPSA
from .optimizerlib import ParametrizedMetaModel

try:
    from .optimizerlib import ParametrizedBO
except ImportError:
    pass  # bayes_opt not available
from .optimizerlib import ParametrizedCMA
from .optimizerlib import EMNA
from .optimizerlib import Chaining
from .optimizerlib import NoisySplit
from .optimizerlib import ConfPortfolio
from .optimizerlib import ConfPSO
from .optimizerlib import ConfSplitOptimizer
from .optimizerlib import BayesOptim
from .differentialevolution import DifferentialEvolution
from .es import EvolutionStrategy
from .recastlib import NonObjectOptimizer
from .recastlib import Pymoo
from .oneshot import RandomSearchMaker
from .oneshot import SamplingSearch


__all__ = [
    "ParametrizedOnePlusOne",
    "ParametrizedCMA",
    "ParametrizedBO",
    "ParametrizedTBPSA",
    "ParametrizedMetaModel",
    "DifferentialEvolution",
    "EvolutionStrategy",
    "NonObjectOptimizer",
    "Pymoo",
    "RandomSearchMaker",
    "SamplingSearch",
    "Chaining",
    "EMNA",
    "NoisySplit",
    "ConfPortfolio",
    "ConfPSO",
    "ConfSplitOptimizer",
    "BayesOptim",
]



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from pathlib import Path
import numpy as np

# pylint: disable=reimported,redefined-outer-name,unused-variable,unsubscriptable-object, unused-argument
# pylint: disable=import-outside-toplevel


def test_concrete_model_example() -> None:
    # DOC_CONCRETE_0
    import pyomo.environ as pyomo

    def square(m):
        return pyomo.quicksum((m.x[i] - 0.5) ** 2 for i in m.x)

    model = pyomo.ConcreteModel()
    model.x = pyomo.Var([0, 1], domain=pyomo.Reals)
    model.obj = pyomo.Objective(rule=square)
    model.Constraint1 = pyomo.Constraint(rule=lambda m: m.x[0] >= 1)
    model.Constraint2 = pyomo.Constraint(rule=lambda m: m.x[1] >= 0.8)
    # DOC_CONCRETE_1

    # DOC_CONCRETE_10
    import nevergrad as ng
    import nevergrad.functions.pyomo as ng_pyomo

    # DOC_CONCRETE_11

    # DOC_CONCRETE_100
    func = ng_pyomo.Pyomo(model)
    optimizer = ng.optimizers.OnePlusOne(parametrization=func.parametrization, budget=100)
    recommendation = optimizer.minimize(func.function)
    # DOC_CONCRETE_101

    # DOC_CONCRETE_1000
    print(recommendation.kwargs["x[0]"])
    print(recommendation.kwargs["x[1]"])
    # DOC_CONCRETE_1001

    np.testing.assert_almost_equal(recommendation.kwargs["x[0]"], 1.0, decimal=1)
    np.testing.assert_almost_equal(recommendation.kwargs["x[1]"], 0.8, decimal=1)


def test_abstract_model_example() -> None:
    import pyomo.environ as pyomo

    def square(m):
        return pyomo.quicksum((m.x[i] - 0.5) ** 2 for i in m.x)

    abstract_model = pyomo.AbstractModel()
    abstract_model.F = pyomo.Set()
    abstract_model.Xmin = pyomo.Param(abstract_model.F, within=pyomo.Reals, default=0.0)
    abstract_model.x = pyomo.Var(abstract_model.F, within=pyomo.Reals)
    abstract_model.constraints = pyomo.Constraint(abstract_model.F, rule=lambda m, i: m.x[i] >= m.Xmin[i])
    abstract_model.obj = pyomo.Objective(rule=square)

    import nevergrad as ng
    import nevergrad.functions.pyomo as ng_pyomo

    # Load the values of the parameters from external file
    data_path = str(Path(__file__).with_name("test_model_1.dat"))

    # DOC_ABSTRACT_100
    data = pyomo.DataPortal()
    data.load(filename=data_path, model=abstract_model)
    model = abstract_model.create_instance(data)
    # DOC_ABSTRACT_101

    func = ng_pyomo.Pyomo(model)
    optimizer = ng.optimizers.OnePlusOne(parametrization=func.parametrization, budget=200)
    recommendation = optimizer.minimize(func.function)

    np.testing.assert_almost_equal(recommendation.kwargs['x["New York"]'], model.Xmin["New York"], decimal=1)
    np.testing.assert_almost_equal(
        recommendation.kwargs['x["Hong Kong"]'], model.Xmin["Hong Kong"], decimal=1
    )



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import numpy as np
import nevergrad as ng
from ..base import ExperimentFunction


class UnitCommitmentProblem(ExperimentFunction):
    """Model that uses conventional implementation for semi-continuous variables
    The model is adopted from Pyomo model 1: Conventional implementation for semi-continuous variables
    (https://jckantor.github.io/ND-Pyomo-Cookbook/04.06-Unit-Commitment.html)
    The constraints are added to the objective with a heavy penalty for violation.

    Parameters
    ----------
    num_timepoints: int
        number of time points
    num_generators: int
        number of generators
    penalty_weight: float
        weight to penalize for violation of constraints
    """

    def __init__(
        self,
        problem_name: str = "semi-continuous",
        num_timepoints: int = 13,
        num_generators: int = 3,
        penalty_weight: float = 10000,
    ) -> None:
        if problem_name not in ["semi-continuous"]:
            raise NotImplementedError

        # Demand for certain time period
        self.num_timepoints = num_timepoints
        self.demands = np.random.uniform(low=100, high=200, size=(self.num_timepoints,))

        # Generating units
        self.num_generators = num_generators
        self.p_max = 2 * max(self.demands) / self.num_generators
        self.p_min = 0.6 * self.p_max

        # Cost
        self.cost_a = np.clip(np.random.randn(self.num_generators) * 0.2 + 0.5, a_min=0, a_max=None)
        self.cost_b = np.random.uniform(low=0, high=10, size=(self.num_generators,))

        self.penalty_weight = penalty_weight

        param_operational_output = ng.p.Array(shape=(self.num_generators, self.num_timepoints)).set_bounds(
            0, self.p_max
        )
        # param_operational_states = ng.p.Array(shape=(self.num_generators, self.num_timepoints)).set_bounds(0, 1).set_integer_casting() #
        param_operational_states = ng.p.Choice([0, 1], repetitions=self.num_timepoints * self.num_generators)
        instru = ng.p.Instrumentation(
            operational_output=param_operational_output,
            operational_states=param_operational_states,
        ).set_name("")
        super().__init__(self.unit_commitment_obj_with_penalization, instru)

    def unit_commitment_obj_with_penalization(self, operational_output, operational_states):
        operational_states = np.array(operational_states).reshape(self.num_generators, self.num_timepoints)
        demand_penalty, lb_penalty, ub_penalty = 0, 0, 0
        # From demand constraint
        demand_penalty = np.sum(np.abs(np.sum(operational_output, axis=0) - self.demands))
        # From semi_continuous_constraints
        lb_penalty = np.sum(
            np.clip(self.p_min * operational_states - operational_output, 0, a_max=None),
            axis=None,
        )
        ub_penalty = np.sum(
            np.clip(operational_output - self.p_max * operational_states, 0, a_max=None),
            axis=None,
        )
        # Running cost
        running_cost = np.sum(
            np.sum(operational_output, axis=1) * self.cost_a
            + np.sum(operational_states, axis=1) * self.cost_b
        )
        ret = running_cost + (demand_penalty + lb_penalty + ub_penalty) * self.penalty_weight
        return ret



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# import os
import sys
import time
import random
import itertools
import contextlib
import typing as tp
from pathlib import Path
import numpy as np
import pytest
from nevergrad.common import testing
from . import transforms as trans
from . import parameter as p
from . import utils
from . import helpers


def test_temporary_directory_copy() -> None:
    filepath = Path(__file__)
    with utils.TemporaryDirectoryCopy(filepath.parent) as cpath:
        assert cpath.exists()
        assert (cpath / filepath.name).exists()
    assert not cpath.exists()


def test_command_function() -> None:
    command = f"{sys.executable} -m nevergrad.parametrization.test_utils".split()
    word = "testblublu12"
    output = utils.CommandFunction(command)(word)
    # if os.environ.get("CIRCLECI", False):
    #     raise SkipTest("Failing in CircleCI")  # TODO investigate why
    assert output is not None
    assert word in output, f'Missing word "{word}" in output:\n{output}'
    try:
        with contextlib.redirect_stderr(sys.stdout):
            output = utils.CommandFunction(command, verbose=True)(error=True)
    except utils.FailedJobError as e:
        words = "Too bad"
        assert words in str(e), f'Missing word "{words}" in output:\n\n{e}'
    else:
        raise AssertionError("An error should have been raised")


@testing.parametrized(
    scalar=(False, p.Scalar(), ("",)),
    v_scalar=(True, p.Scalar(), ("",)),
    tuple_=(False, p.Tuple(p.Scalar(), p.Array(shape=(2,))), ("", "0", "1")),
    v_tuple_=(True, p.Tuple(p.Scalar(), p.Array(shape=(2,))), ("0", "1")),
    instrumentation=(False, p.Instrumentation(p.Scalar(), y=p.Scalar()), ("", "0", "y")),
    instrumentation_v=(True, p.Instrumentation(p.Scalar(), y=p.Scalar()), ("0", "y")),
    choice=(False, p.Choice([p.Scalar(), "blublu"]), ("", "choices", "choices.0", "choices.1", "indices")),
    v_choice=(True, p.Choice([p.Scalar(), "blublu"]), ("", "choices.0", "indices")),
    tuple_choice_dict=(
        False,
        p.Tuple(p.Choice([p.Dict(x=p.Scalar(), y=12), p.Scalar()])),
        ("", "0", "0.choices", "0.choices.0", "0.choices.0.x", "0.choices.0.y", "0.choices.1", "0.indices"),
    ),
    v_tuple_choice_dict=(
        True,
        p.Tuple(p.Choice([p.Dict(x=p.Scalar(), y=12), p.Scalar()])),
        ("0", "0.choices.0.x", "0.choices.1", "0.indices"),
    ),
)
def test_flatten(no_container: bool, param: p.Parameter, keys: tp.Iterable[str]) -> None:
    flat = dict(helpers.flatten(param, with_containers=not no_container))
    assert set(flat) == set(keys), f"Unexpected flattened parameter: {flat}"


def test_function_info() -> None:
    info = utils.FunctionInfo(deterministic=False)
    assert repr(info) == "FunctionInfo(deterministic=False,metrizable=True,proxy=False)"


@testing.parametrized(
    # updating this tests requires checking manually through prints
    # that everything works as intended
    v_tuple_choice_dict=(
        p.Tuple(p.Choice([p.Dict(x=p.Scalar(), y=12), p.Scalar()])),
        ["0.choices.0.x", "0.choices.1", "0.indices"],
    ),
    multiple=(
        p.Instrumentation(
            p.Scalar(init=12, lower=12, upper=12.01),
            x=p.Choice([3, p.Log(lower=0.01, upper=0.1)]),
            z=p.Array(init=[12, 12]).set_bounds(lower=12, upper=15),
            y=p.Array(init=[1, 1]),
        ),
        ["0", "x.choices.1", "x.indices", "y", "z"],
    ),
)
def test_split_as_data_parameters(param: p.Parameter, names: tp.List[str]) -> None:
    # new version
    output = helpers.flatten(param)
    assert [x[0] for x in output if isinstance(x[1], p.Data)] == names
    # legacy
    output2 = split_as_data_parameters(param)
    assert [x[0] for x in output2] == names


@testing.parametrized(
    order_0=(0, ("", "choices.0.x", "choices.1", "indices")),
    order_1=(1, ("", "choices.0.x", "choices.1", "indices", "choices.1#sigma", "choices.0.x#sigma")),
    order_2=(
        2,
        (
            "",
            "choices.0.x",
            "choices.1",
            "indices",
            "choices.1#sigma",
            "choices.0.x#sigma",
            "choices.1#sigma#sigma",
        ),
    ),
    order_3=(
        3,
        (
            "",
            "choices.0.x",
            "choices.1",
            "indices",
            "choices.1#sigma",
            "choices.0.x#sigma",
            "choices.1#sigma#sigma",
        ),
    ),
)
def test_flatten_order(order: int, keys: tp.Iterable[str]) -> None:
    param = p.Choice([p.Dict(x=p.Scalar(), y=12), p.Scalar().sigma.set_mutation(sigma=p.Scalar())])
    flat = dict(helpers.flatten(param, with_containers=False, order=order))
    assert set(flat) == set(keys), f"Unexpected flattened parameter: {flat}"


@testing.parametrized(
    true=(True, 0.0),
    false=(False, 1.0),
    np_true=(np.bool_(True), 0.0),
    np_false=(np.bool_(False), 1.0),
    pos=(0.7, 0.0),
    neg=(-0.7, 0.7),
    np_pos=(np.float_(0.7), 0.0),
    np_neg=(np.float_(-0.7), 0.7),
)
def test_float_penalty(value: tp.Any, expected: float) -> None:
    assert utils.float_penalty(value) == expected


# # # OLD FUNCTION SERVING AS CHECK FOR DATA ORDER # # #


# pylint: disable=too-many-locals
def split_as_data_parameters(
    parameter: p.Parameter,
) -> tp.List[tp.Tuple[str, p.Data]]:
    """List all the instances involved as parameter (not as subparameter/
    endogeneous parameter) ordered as in standardized data space

    Parameter
    ---------
    parameter: Parameter
        the parameter to split

    Returns
    -------
    list
        the list and subparameters ordered as in data space

    Note
    ----
    This function is experimental, its output will probably evolve before converging.
    """
    err_msg = (
        f"Could not figure out the data order for: {parameter} "
        "(please open an issue on nevergrad github repository)"
    )
    copied = parameter.copy()
    ref = parameter.copy()
    flatp, flatc, flatref = (
        {x: y for x, y in helpers.flatten(pa) if isinstance(y, p.Data)} for pa in (parameter, copied, ref)
    )
    keys = list(flatp.keys())
    random.shuffle(keys)  # makes it safer to test!
    # remove transforms for both ref and copied parameters and set index
    for k, key in enumerate(keys):
        for not_ref, flat in enumerate((flatref, flatc)):
            param = flat[key]
            param._layers = param._layers[:1]  # force remove the bound to avoid weird clipping etc
            param.set_mutation(sigma=1.0)  # force mutation sigma to 1 to avoid rounding
            if not_ref:
                param.set_standardized_data(k * np.ones((param.dimension)))
    # analyze results
    data = copied.get_standardized_data(reference=ref)
    order: tp.List[int] = []
    for val, _ in itertools.groupby(data):
        num = int(np.round(val))
        if num in order:
            if order[-1] != num:
                raise RuntimeError(err_msg)
        else:
            order.append(num)
    if len(order) != len(flatp):
        raise RuntimeError(err_msg)
    # create output and check it
    ordered_keys = [keys[num] for num in order]
    ordered_arrays = [(k, flatp[k]) for k in ordered_keys]
    # print(f"DEBUGGING:\nkeys={keys}\ndata={data}\norder={order}\nordered_key={ordered_keys}")
    if sum(pa.dimension for _, pa in ordered_arrays) != parameter.dimension:
        raise RuntimeError(err_msg)
    return ordered_arrays


def test_normalizer_backward() -> None:
    ref = p.Instrumentation(
        p.Array(shape=(1, 2)).set_bounds(-12, 12, method="arctan"),
        p.Array(shape=(2,)).set_bounds(-12, 12, full_range_sampling=False),
        lr=p.Log(lower=0.001, upper=1000),
        stuff=p.Scalar(lower=-1, upper=2),
        unbounded=p.Scalar(lower=-1, init=0.0),
        value=p.Scalar(),
        letter=p.Choice("abc"),
    )
    # make sure the order is preserved using legacy split method
    expected = [x[1] for x in split_as_data_parameters(ref)]
    assert helpers.list_data(ref) == expected
    # check the bounds
    param = ref.spawn_child()
    scaler = helpers.Normalizer(param, only_sampling=True, unbounded_transform=trans.Affine(1, 0))
    assert not scaler.fully_bounded
    # setting
    output = scaler.backward([1.0] * param.dimension)
    param.set_standardized_data(output)
    (array1, array2), values = param.value
    np.testing.assert_array_almost_equal(array1, [[12, 12]])
    np.testing.assert_array_almost_equal(array2, [1, 1])
    assert values["stuff"] == 2
    assert values["unbounded"] == 1
    assert values["value"] == 1
    assert values["lr"] == pytest.approx(1000)
    # again, on the middle point
    output = scaler.backward([0] * param.dimension)
    param.set_standardized_data(output)
    assert param.value[1]["lr"] == pytest.approx(1.0)
    assert param.value[1]["stuff"] == pytest.approx(0.5)


def test_normalizer_forward() -> None:
    ref = p.Tuple(
        p.Scalar(init=0), p.Scalar(init=1), p.Scalar(lower=-1, upper=3, init=0), p.Scalar(lower=-1, upper=1)
    )
    scaler = helpers.Normalizer(ref)
    assert not scaler.fully_bounded
    out = scaler.forward(ref.get_standardized_data(reference=ref))
    np.testing.assert_almost_equal(out, [0.5, 0.5, 0.25, 0.5])
    param = ref.spawn_child()
    param.value = (0, 100, -1, 1)
    data = param.get_standardized_data(reference=ref)
    out = scaler.forward(data)
    np.testing.assert_almost_equal(out, [0.5, 0.9967849, 0, 1])
    back = scaler.backward(out)
    np.testing.assert_almost_equal(back, data)  # shouuld be bijective


@testing.parametrized(
    set_bounds=(True, p.Array(shape=(1, 2)).set_bounds(-12, 12, method="arctan", full_range_sampling=False)),
    log=(True, p.Log(lower=0.001, upper=1000)),
    partial=(False, p.Scalar(lower=-12, init=0)),
    none=(False, p.Scalar()),
    full_tuple=(True, p.Tuple(p.Log(lower=0.01, upper=1), p.Scalar(lower=0, upper=1))),
    partial_tuple=(False, p.Tuple(p.Log(lower=0.01, upper=1), p.Scalar(lower=0, init=1))),
)
def test_scaler_fully_bounded(expected: float, param: p.Parameter) -> None:
    scaler = helpers.Normalizer(param)
    assert scaler.fully_bounded == expected


# # # END OF CHECK # # #


def do_nothing(*args: tp.Any, **kwargs: tp.Any) -> int:
    print("my args", args, flush=True)
    print("my kwargs", kwargs, flush=True)
    if "sleep" in kwargs:
        print("Waiting", flush=True)
        time.sleep(int(kwargs["sleep"]))
    if kwargs.get("error", False):
        print("Raising", flush=True)
        raise ValueError("Too bad")
    print("Finishing", flush=True)
    return 12


if __name__ == "__main__":
    c_args, c_kwargs = [], {}  # oversimplisitic parser
    for argv in sys.argv[1:]:
        if "=" in argv:
            key_, val_ = argv.split("=")
            c_kwargs[key_.strip("-")] = val_
        else:
            c_args.append(argv)
    do_nothing(*c_args, **c_kwargs)



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from collections import OrderedDict
import numpy as np
import nevergrad.common.typing as tp
from . import utils
from . import core


D = tp.TypeVar("D", bound="Container")


class Container(core.Parameter):
    """Parameter which can hold other parameters.
    This abstract implementation is based on a dictionary.

    Parameters
    ----------
    **parameters: Any
        the objects or Parameter which will provide values for the dict

    Note
    ----
    This is the base structure for all container Parameters, and it is
    used to hold the internal/model parameters for all Parameter classes.
    """

    def __init__(self, **parameters: tp.Any) -> None:
        super().__init__()
        self._subobjects = utils.Subobjects(self, base=core.Parameter, attribute="_content")
        self._content: tp.Dict[tp.Any, core.Parameter] = {
            k: core.as_parameter(p) for k, p in parameters.items()
        }
        self._sizes: tp.Optional[tp.Dict[str, int]] = None
        self._sanity_check(list(self._content.values()))
        self._ignore_in_repr: tp.Dict[
            str, str
        ] = {}  # hacky undocumented way to bypass boring representations

    @property
    def dimension(self) -> int:
        return sum(x.dimension for x in self._content.values())

    def _sanity_check(self, parameters: tp.List[core.Parameter]) -> None:
        """Check that all parameters are different"""
        # TODO: this is first order, in practice we would need to test all the different
        # parameter levels together
        if parameters:
            assert all(isinstance(p, core.Parameter) for p in parameters)
            ids = {id(p) for p in parameters}
            if len(ids) != len(parameters):
                raise ValueError("Don't repeat twice the same parameter")

    def __getitem__(self, name: tp.Any) -> core.Parameter:
        return self._content[name]

    def __setitem__(self, name: tp.Any, value: tp.Any) -> None:
        self._sizes = None
        self._content[name] = core.as_parameter(value)
        if self._random_state is not None:
            self._content[name].random_state = self.random_state
        self._sanity_check(list(self._content.values()))

    def __len__(self) -> int:
        return len(self._content)

    def _get_parameters_str(self) -> str:
        raise NotImplementedError

    def _get_name(self) -> str:
        return f"{self.__class__.__name__}({self._get_parameters_str()})"

    def get_value_hash(self) -> tp.Hashable:
        return tuple(sorted((x, y.get_value_hash()) for x, y in self._content.items()))

    def _internal_get_standardized_data(self: D, reference: D) -> np.ndarray:
        data = {k: self[k].get_standardized_data(reference=p) for k, p in reference._content.items()}
        if self._sizes is None:
            self._sizes = OrderedDict(sorted((x, y.size) for x, y in data.items()))
        assert self._sizes is not None
        data_list = [data[k] for k in self._sizes]
        if not data_list:
            return np.array([])
        return data_list[0] if len(data_list) == 1 else np.concatenate(data_list)  # type: ignore

    def _internal_set_standardized_data(self: D, data: np.ndarray, reference: D) -> None:
        if self._sizes is None:
            self.get_standardized_data(reference=self)
        assert self._sizes is not None
        if data.size != sum(v for v in self._sizes.values()):
            raise ValueError(
                f"Unexpected shape {data.shape} for {self} with dimension {self.dimension}:\n{data}"
            )
        data = data.ravel()
        start, end = 0, 0
        for name, size in self._sizes.items():
            end = start + size
            self._content[name].set_standardized_data(data[start:end], reference=reference[name])
            start = end
        assert end == len(data), f"Finished at {end} but expected {len(data)}"

    def _layered_sample(self: D) -> D:
        child = self.spawn_child()
        child._content = {k: p.sample() for k, p in self._content.items()}
        return child

    def _layered_recombine(self: D, *others: D) -> None:  # type: ignore
        pass  # nothing specific here, already propagated to sub-objects (content)


class Dict(Container):
    """Dictionary-valued parameter. This Parameter can contain other Parameters,
    its value is a dict, with keys the ones provided as input, and corresponding values are
    either directly the provided values if they are not Parameter instances, or the value of those
    Parameters. It also implements a getter to access the Parameters directly if need be.

    Parameters
    ----------
    **parameters: Any
        the objects or Parameter which will provide values for the dict

    Note
    ----
    This is the base structure for all container Parameters, and it is
    used to hold the internal/model parameters for all Parameter classes.
    """

    value: core.ValueProperty[tp.Dict[str, tp.Any], tp.Dict[str, tp.Any]] = core.ValueProperty()

    def __iter__(self) -> tp.Iterator[str]:
        return iter(self.keys())

    def keys(self) -> tp.KeysView[str]:
        return self._content.keys()

    def items(self) -> tp.ItemsView[str, core.Parameter]:
        return self._content.items()

    def values(self) -> tp.ValuesView[core.Parameter]:
        return self._content.values()

    def _layered_get_value(self) -> tp.Dict[str, tp.Any]:
        return {k: p.value for k, p in self.items()}

    def _layered_set_value(self, value: tp.Dict[str, tp.Any]) -> None:
        cls = self.__class__.__name__
        if not isinstance(value, dict):
            raise TypeError(f"{cls} value must be a dict, got: {value}\nCurrent value: {self.value}")
        if set(value) != set(self):
            raise ValueError(
                f"Got input keys {set(value)} for {cls} but expected {set(self._content)}\nCurrent value: {self.value}"
            )
        for key, val in value.items():
            self._content[key].value = val

    def _get_parameters_str(self) -> str:
        params = sorted(
            (k, p.name) for k, p in self.items() if p.name != self._ignore_in_repr.get(k, "#ignoredrepr#")
        )
        return ",".join(f"{k}={n}" for k, n in params)


class Tuple(Container):
    """Tuple-valued parameter. This Parameter can contain other Parameters,
    its value is tuple which values are either directly the provided values
    if they are not Parameter instances, or the value of those Parameters.
    It also implements a getter to access the Parameters directly if need be.

    Parameters
    ----------
    **parameters: Any
        the objects or Parameter which will provide values for the tuple

    Note
    ----
    This is the base structure for all container Parameters, and it is
    used to hold the subparameters for all Parameter classes.
    """

    def __init__(self, *parameters: tp.Any) -> None:
        super().__init__()
        self._content.update({k: core.as_parameter(p) for k, p in enumerate(parameters)})
        self._sanity_check(list(self._content.values()))

    def _get_parameters_str(self) -> str:
        return ",".join(f"{param.name}" for param in self)

    def __iter__(self) -> tp.Iterator[core.Parameter]:
        return (self._content[k] for k in range(len(self)))

    value: core.ValueProperty[tp.Tuple[tp.Any, ...], tp.Tuple[tp.Any, ...]] = core.ValueProperty()

    def _layered_get_value(self) -> tp.Tuple[tp.Any, ...]:
        return tuple(p.value for p in self)

    def _layered_set_value(self, value: tp.Tuple[tp.Any, ...]) -> None:
        if not isinstance(value, tuple) or not len(value) == len(self):
            cls = self.__class__.__name__
            raise ValueError(
                f"{cls} value must be a tuple of size {len(self)}, got: {value}.\nCurrent value: {self.value}"
            )
        for k, val in enumerate(value):
            core.as_parameter(self[k]).value = val


class Instrumentation(Tuple):
    """Container of parameters available through `args` and `kwargs` attributes.
    The parameter provided as input are used to provide values for
    an `arg` tuple and a `kwargs` dict.
    `value` attribue returns `(args, kwargs)`, but each can be independantly
    accessed through the `args` and `kwargs` properties.

    Parameters
    ----------
    *args
         values or Parameters to be used to fill the tuple of args
    *kwargs
         values or Parameters to be used to fill the dict of kwargs

    Note
    ----
    When used in conjonction with the "minimize" method of an optimizer,
    functions call use `func(*param.args, **param.kwargs)` instead of
    `func(param.value)`. This is for simplifying the parametrization of
    multiparameter functions.
    """

    def __init__(self, *args: tp.Any, **kwargs: tp.Any) -> None:
        super().__init__(Tuple(*args), Dict(**kwargs))
        self._sanity_check(list(self[0]._content.values()) + list(self[1]._content.values()))  # type: ignore

    @property
    def args(self) -> tp.Tuple[tp.Any, ...]:
        return self.value[0]

    @property
    def kwargs(self) -> tp.Dict[str, tp.Any]:
        return self.value[1]

    value: core.ValueProperty[tp.ArgsKwargs, tp.ArgsKwargs] = core.ValueProperty()  # type: ignore



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import re
import os
import warnings
import platform
import unittest
import inspect
import contextlib
from pathlib import Path
import typing as tp
import numpy as np
from . import errors

try:
    import pytest
except ImportError:
    pass  # makes most of this module usable without pytest


@contextlib.contextmanager
def suppress_nevergrad_warnings() -> tp.Iterator[None]:
    with warnings.catch_warnings():
        # tests do not need to be efficient
        warnings.simplefilter("ignore", category=errors.NevergradWarning)
        yield


def assert_set_equal(
    estimate: tp.Iterable[tp.Any], reference: tp.Iterable[tp.Any], err_msg: str = ""
) -> None:
    """Asserts that both sets are equals, with comprehensive error message.
    This function should only be used in tests.
    Parameters
    ----------
    estimate: iterable
        sequence of elements to compare with the reference set of elements
    reference: iterable
        reference sequence of elements
    """
    estimate, reference = (set(x) for x in [estimate, reference])
    elements = [("additional", estimate - reference), ("missing", reference - estimate)]
    messages = ["  - {} element(s): {}.".format(name, s) for (name, s) in elements if s]
    if messages:
        messages = ([err_msg] if err_msg else []) + ["Sets are not equal:"] + messages
        raise AssertionError("\n".join(messages))


def printed_assert_equal(actual: tp.Any, desired: tp.Any, err_msg: str = "") -> None:
    try:
        np.testing.assert_equal(actual, desired, err_msg=err_msg)
    except AssertionError as e:
        print("\n" + "# " * 12 + "DEBUG MESSAGE " + "# " * 12)
        print(f"Expected: {desired}\nbut got:  {actual}")
        raise e


def assert_markdown_links_not_broken(folder: tp.Union[str, Path]) -> None:
    """Asserts that all relative hyperlinks are valid in markdown files of the folder
    and its subfolders.

    Note
    ----
    http hyperlinks are not tested.
    """
    links = _get_all_markdown_links(folder)
    broken = [l for l in links if not l.exists()]
    if broken:
        text = "\n - ".join([str(l) for l in broken])
        raise AssertionError(f"Broken markdown links:\n - {text}")


class _MarkdownLink:
    """Handle to a markdown link, for easy existence test and printing
    (external links are not tested)
    """

    def __init__(self, folder: Path, filepath: Path, string: str, link: str) -> None:
        self._folder = folder
        self._filepath = filepath
        self._string = string
        self._link = link

    def exists(self) -> bool:
        if self._link.startswith("http"):  # consider it exists
            return True
        fullpath = self._folder / self._filepath.parent / self._link
        return fullpath.exists()

    def __repr__(self) -> str:
        return f"{self._link} ({self._string}) from file {self._filepath}"


def _get_all_markdown_links(folder: tp.Union[str, Path]) -> tp.List[_MarkdownLink]:
    """Returns a list of all existing markdown links"""
    pattern = re.compile(r"\[(?P<string>.+?)\]\((?P<link>\S+?)\)")
    folder = Path(folder).expanduser().absolute()
    links = []
    for rfilepath in folder.glob("**/*.md"):
        if ("/site-packages/" if os.name != "nt" else "\\site-packages\\") not in str(rfilepath):
            filepath = folder / rfilepath
            with filepath.open("r") as f:
                text = f.read()
            for match in pattern.finditer(text):
                links.append(_MarkdownLink(folder, rfilepath, match.group("string"), match.group("link")))
    return links


class parametrized:
    """Simplified decorator API for specifying named parametrized test with pytests
    (like with old "genty" package)
    See example of use in test_testing

    Parameters
    ----------
    **kwargs:
        name of the argument is converted as id of the experiments, and the provided tuple
        contains a value for each of the arguments of the underlying function (in the definition order).
    """

    def __init__(self, **kwargs: tp.Tuple[tp.Any, ...]):
        self.ids = sorted(kwargs)
        self.params = tuple(kwargs[name] for name in self.ids)
        assert self.params
        self.num_params = len(self.params[0])
        assert all(isinstance(p, (tuple, list)) for p in self.params)
        assert all(self.num_params == len(p) for p in self.params[1:])

    def __call__(self, func: tp.Callable[..., None]) -> tp.Any:  # type is lost here :(
        names = list(inspect.signature(func).parameters.keys())
        assert len(names) == self.num_params, f"Parameter names: {names}"
        return pytest.mark.parametrize(
            ",".join(names), self.params if self.num_params > 1 else [p[0] for p in self.params], ids=self.ids
        )(func)


@contextlib.contextmanager
def skip_error_on_systems(error_type: tp.Type[Exception], systems: tp.Iterable[str]) -> tp.Iterator[None]:
    """Context manager for skipping a test upon a specific error on specific systems
    This is mostly used to skip some tests for features which are incompatible with Windows
    Eg. of systems (mind the capitalized first letter): Darwin, Windows
    """
    try:
        yield
    except error_type as e:
        system = platform.system()
        if system in systems:
            raise unittest.SkipTest(f"Skipping on system {system}")
        if systems:  # only print if the context is actually active for some system
            print(f'This is system "{system}" (should it be skipped for the test?)')
        raise e



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import warnings
from pathlib import Path
import numpy as np
import nevergrad.common.typing as tp
from nevergrad.common import testing
import nevergrad as ng
from . import optimizerlib
from . import experimentalvariants as xpvariants
from . import base
from . import utils
from . import callbacks


class CounterFunction:
    def __init__(self) -> None:
        self.count = 0

    def __call__(self, value: tp.ArrayLike) -> float:
        assert len(value) == 1
        self.count += 1
        return float(value[0] - 1) ** 2


class LoggingOptimizer(base.Optimizer):
    def __init__(self, num_workers: int = 1) -> None:
        super().__init__(parametrization=1, budget=5, num_workers=num_workers)
        self.logs: tp.List[str] = []

    def _internal_ask(self) -> tp.ArrayLike:
        self.logs.append(f"s{self._num_ask}")  # s for suggest
        return np.array((float(self._num_ask),))

    # pylint: disable=unused-argument
    def _internal_tell(self, x: tp.ArrayLike, loss: float) -> None:
        self.logs.append(f"u{int(x[0])}")  # u for update


@testing.parametrized(
    w1_batch=(1, True, ["s0", "u0", "s1", "u1", "s2", "u2", "s3", "u3", "s4", "u4"]),
    w1_steady=(
        1,
        False,
        ["s0", "u0", "s1", "u1", "s2", "u2", "s3", "u3", "s4", "u4"],
    ),  # no difference (normal, since worker=1)
    w3_batch=(3, True, ["s0", "s1", "s2", "u0", "u1", "u2", "s3", "s4", "u3", "u4"]),
    w3_steady=(
        3,
        False,
        ["s0", "s1", "s2", "u0", "u1", "u2", "s3", "s4", "u3", "u4"],
    ),  # not really steady TODO change this behavior
    # w3_steady=(3, False, ['s0', 's1', 's2', 'u0', 's3', 'u1', 's4', 'u2', 'u3', 'u4']),  # This is what we would like
)
def test_batch_and_steady_optimization(
    num_workers: int, batch_mode: bool, expected: tp.List[tp.Tuple[str, float]]
) -> None:
    # tests the suggestion (s) and update (u) patterns
    # the w3_steady is unexpected. It is designed to be efficient with a non-sequential executor, but because
    # of it, it is acting like batch mode when sequential...
    optim = LoggingOptimizer(num_workers=num_workers)
    func = CounterFunction()
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        optim.minimize(func, verbosity=2, batch_mode=batch_mode)
    testing.printed_assert_equal(optim.logs, expected)


@testing.parametrized(
    int_val=(3, False),
    bool_val=(True, False),
    int32_val=(np.int32(3), False),
    int64_val=(np.int64(3), False),
    float64_val=(np.float64(3), False),
    float32_val=(np.float32(3), False),
    list_val=([3, 5], False),
    complex_val=(1j, True),
    object_val=(object(), True),
)
def test_tell_types(value: tp.Any, error: bool) -> None:
    optim = LoggingOptimizer(num_workers=1)
    x = optim.ask()
    if error:
        np.testing.assert_raises(TypeError, optim.tell, x, value)
    else:
        optim.tell(x, value)
    x = optim.ask()
    if error:
        np.testing.assert_raises(TypeError, optim.tell, x, value)
    else:
        optim.tell(x, value, [3.0, 5.0, -1.0])  # Standard penalty function
    x = optim.ask()
    if error:
        np.testing.assert_raises(TypeError, optim.tell, x, value)
    else:
        optim.tell(x, value, [3.0, 5.0, -1.0], [1.0, 1.0, 1.0, 1.0, 1.0, 1.0])  # User-chosen penalty function

    # def tell(self, candidate: p.Parameter, loss: tp.Loss, constraint_violation: tp.Optional[tp.Loss] = None, penalty_style: tp.Optional[str]) -> None:


def test_base_optimizer() -> None:
    zeroptim = xpvariants.Zero(parametrization=2, budget=4, num_workers=1)
    zeroptim.parametrization.function.deterministic = False
    assert not zeroptim.parametrization.function.deterministic
    representation = repr(zeroptim)
    expected = "parametrization=Array{(2,)}"
    assert expected in representation, f"Unexpected representation: {representation}"
    np.testing.assert_equal(zeroptim.ask().value, [0, 0])
    zeroptim.tell(zeroptim.parametrization.spawn_child().set_standardized_data([0.0, 0]), 0)
    zeroptim.tell(zeroptim.parametrization.spawn_child().set_standardized_data([1.0, 1]), 1)
    np.testing.assert_equal(zeroptim.provide_recommendation().value, [0, 0])
    # check that the best value is updated if a second evaluation is not as good
    zeroptim.tell(zeroptim.parametrization.spawn_child().set_standardized_data([0.0, 0]), 10)
    zeroptim.tell(zeroptim.parametrization.spawn_child().set_standardized_data([1.0, 1]), 1)
    np.testing.assert_equal(zeroptim.provide_recommendation().value, [1, 1])
    np.testing.assert_equal(zeroptim._num_ask, 1)
    # check suggest
    zeroptim.suggest([12, 12])
    np.testing.assert_array_equal(zeroptim.ask().args[0], [12, 12])
    np.testing.assert_array_equal(zeroptim.ask().args[0], [0, 0])


def test_optimize_and_dump(tmp_path: Path) -> None:
    optimizer = optimizerlib.OnePlusOne(parametrization=1, budget=100, num_workers=5)
    optimizer.register_callback(
        "tell", callbacks.OptimizationPrinter(print_interval_tells=10, print_interval_seconds=0.1)
    )
    func = CounterFunction()
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        result = optimizer.minimize(func, verbosity=2)
    np.testing.assert_almost_equal(result.value[0], 1, decimal=2)
    np.testing.assert_equal(func.count, 100)
    # pickling
    filepath = tmp_path / "dump_test.pkl"
    optimizer.dump(filepath)
    optimizer2 = optimizerlib.OnePlusOne.load(filepath)
    np.testing.assert_almost_equal(optimizer2.provide_recommendation().value[0], 1, decimal=2)


def test_compare() -> None:
    optimizer = optimizerlib.OnePlusOne(parametrization=2, budget=600, num_workers=6)
    optimizerlib.addCompare(optimizer)
    for _ in range(100):  # TODO make faster test
        x: tp.List[tp.Any] = []
        for _ in range(6):
            x += [optimizer.ask()]
        winners = sorted(x, key=lambda x_: np.linalg.norm(x_.value - np.array((1.0, 1.0))) ** 2)
        optimizer.compare(winners[:3], winners[3:])  # type: ignore
    result = optimizer.provide_recommendation()
    np.testing.assert_almost_equal(result.value[0], 1.00, decimal=1)


def test_naming() -> None:
    optf = optimizerlib.RandomSearchMaker(stupid=True)
    opt = optf(parametrization=2, budget=4, num_workers=1)
    instru_str = "Array{(2,)}"
    np.testing.assert_equal(
        repr(opt),
        f"Instance of RandomSearchMaker(stupid=True)(parametrization={instru_str}, budget=4, num_workers=1)",
    )
    optf.set_name("BlubluOptimizer", register=True)
    opt = base.registry["BlubluOptimizer"](parametrization=2, budget=4, num_workers=1)
    np.testing.assert_equal(
        repr(opt), f"Instance of BlubluOptimizer(parametrization={instru_str}, budget=4, num_workers=1)"
    )


class MinStorageFunc:
    """Stores the minimum value obtained so far"""

    def __init__(self) -> None:
        self.min_loss = float("inf")

    def __call__(self, score: int) -> float:
        self.min_loss = min(score, self.min_loss)
        return score


def test_recommendation_correct() -> None:
    # Run this several times to debug:
    # pytest nevergrad/optimization/test_base.py::test_recommendation_correct --count=20 --exitfirst
    func = MinStorageFunc()
    choice_size = 20
    param = ng.p.Choice(range(choice_size)).set_name(f"Choice{choice_size}")
    optimizer = optimizerlib.OnePlusOne(parametrization=param, budget=300, num_workers=1)
    recommendation = optimizer.minimize(func)
    assert func.min_loss == recommendation.value


def constant(x: np.ndarray) -> float:  # pylint: disable=unused-argument
    return 12.0


def test_pruning_calls() -> None:
    opt = ng.optimizers.CMA(50, budget=2000)
    # worst case scenario for pruning is constant:
    # it should not keep everything or that will make computation time explode
    opt.minimize(constant)
    assert isinstance(opt.pruning, utils.Pruning)
    assert opt.pruning._num_prunings < 4



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import gym
import numpy as np
from gym.spaces import Box, Tuple, MultiDiscrete


class TupleActionSpace(gym.Env):
    """
    Environment with tuple of discrete spaces at each time step.
    """

    def __init__(self):

        super().__init__()
        self.current_step = 0
        self.num_cells = 71
        self.action_space = Tuple((MultiDiscrete(self.num_cells * [2]), MultiDiscrete(self.num_cells * [6])))
        self.horizon = 168
        self._reward = 0.0

        self.observation_space = Box(
            low=np.array([0, 0, 0]),
            high=np.array([1, 1, 1]),
            dtype=np.float16,
        )

    def _next_observation(self):

        return np.asarray((0.5, 0.5, 0.5), dtype="float16")

    def reward(
        self,
    ) -> float:
        return float(self._reward)

    def _take_action(self, action):
        assert len(action) == 2
        assert len(action[0]) == self.num_cells
        for u in action[1]:
            assert int(u) == u
        self._reward = np.sum(action[1]) - np.sum(action[0])
        return self._reward

    def reset(self):
        return self._next_observation()

    def step(self, action):
        # Execute one time step within the environment
        self._take_action(action)
        obs = self._next_observation()
        self.current_step += 1
        return obs, self.reward(), self.current_step == self.horizon, {}



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import pickle
import pytest
import numpy as np
import nevergrad as ng
from nevergrad.common import testing
from .test_base import CounterFunction
from . import experimentalvariants as xpvariants
from . import utils


def test_value_and_point() -> None:
    param = ng.p.Scalar(init=12.0)
    v = utils.MultiValue(param, 4, reference=param)
    np.testing.assert_equal(v.count, 1)
    v.add_evaluation(3)
    np.testing.assert_equal(v.count, 2)
    np.testing.assert_equal(v.mean, 3.5)
    np.testing.assert_equal(v.square, 12.5)
    np.testing.assert_almost_equal(v.variance, 0.3536, decimal=4)
    assert v.optimistic_confidence_bound < v.pessimistic_confidence_bound
    assert v.get_estimation("optimistic") < v.get_estimation("pessimistic")
    assert v.get_estimation("minimum") == 3
    np.testing.assert_raises(NotImplementedError, v.get_estimation, "blublu")
    repr(v)
    assert v.parameter.value == 12


def test_sequential_executor() -> None:
    func = CounterFunction()
    executor = utils.SequentialExecutor()
    job1 = executor.submit(func, [3])
    np.testing.assert_equal(job1.done(), True)
    np.testing.assert_equal(job1.result(), 4)
    np.testing.assert_equal(func.count, 1)
    job2 = executor.submit(func, [3])
    np.testing.assert_equal(job2.done(), True)
    np.testing.assert_equal(func.count, 1)  # not computed just yet
    job2.result()
    np.testing.assert_equal(func.count, 2)


def test_get_nash() -> None:
    zeroptim = xpvariants.Zero(parametrization=1, budget=4, num_workers=1)
    param = zeroptim.parametrization
    for k in range(4):
        array = (float(k),)
        zeroptim.archive[array] = utils.MultiValue(param, k, reference=param)
        zeroptim.archive[array].count += 4 - k
    nash = utils._get_nash(zeroptim)
    testing.printed_assert_equal(nash, [((2,), 3), ((1,), 4), ((0,), 5)])
    np.random.seed(12)
    output = utils.sample_nash(zeroptim)
    np.testing.assert_equal(output, (2,))


def test_archive() -> None:
    data = [1, 4.5, 12, 0]
    archive: utils.Archive[int] = utils.Archive()
    archive[np.array(data)] = 12
    np.testing.assert_equal(archive[np.array(data)], 12)
    np.testing.assert_equal(archive.get(data), 12)
    np.testing.assert_equal(archive.get([0, 12.0]), None)
    y = np.frombuffer(next(iter(archive.bytesdict.keys())))
    assert data in archive
    np.testing.assert_equal(y, data)
    items = list(archive.items_as_arrays())
    assert isinstance(items[0][0], np.ndarray)
    keys = list(archive.keys_as_arrays())
    assert isinstance(keys[0], np.ndarray)
    repr(archive)
    str(archive)


def test_archive_errors() -> None:
    archive: utils.Archive[float] = utils.Archive()
    archive[[12, 0.0]] = 12.0
    np.testing.assert_raises(AssertionError, archive.__getitem__, [12, 0])  # int instead of float
    np.testing.assert_raises(AssertionError, archive.__getitem__, [[12], [0.0]])  # type: ignore
    np.testing.assert_raises(RuntimeError, archive.keys)
    np.testing.assert_raises(RuntimeError, archive.items)


def test_pruning() -> None:
    param = ng.p.Scalar(init=12.0)
    archive: utils.Archive[utils.MultiValue] = utils.Archive()
    for k in range(3):
        value = utils.MultiValue(param, float(k), reference=param)
        archive[(float(k),)] = value
    value = utils.MultiValue(param, 1.0, reference=param)
    value.add_evaluation(1.0)
    archive[(3.0,)] = value
    # pruning
    pruning = utils.Pruning(min_len=1, max_len=3)
    # 0 is best optimistic and average, and 3 is best pessimistic (variance=0)
    archive = pruning(archive)
    testing.assert_set_equal([x[0] for x in archive.keys_as_arrays()], [0, 3], err_msg=f"Repetition #{k+1}")
    pickle.dumps(archive)  # should be picklable
    # should not change anything this time
    archive2 = pruning(archive)
    testing.assert_set_equal([x[0] for x in archive2.keys_as_arrays()], [0, 3], err_msg=f"Repetition #{k+1}")


@pytest.mark.parametrize(  # type: ignore
    "nw,dimension,expected_min,expected_max",
    [
        (12, 8, 100, 1000),
        (24, 8, 168, 1680),
        (24, 100000, 168, 671),
        (24, 1000000, 168, 504),
    ],
)
def test_pruning_sensible_default(nw: int, dimension: int, expected_min: int, expected_max: int) -> None:
    pruning = utils.Pruning.sensible_default(num_workers=nw, dimension=dimension)
    assert pruning.min_len == expected_min
    assert pruning.max_len == expected_max


def test_uid_queue() -> None:
    uidq = utils.UidQueue()
    for uid in ["a", "b", "c"]:
        uidq.tell(uid)
    for uid in ["a", "b"]:
        assert uidq.ask() == uid
    uidq.tell("b")
    for uid in ["c", "b", "a", "c", "b", "a"]:
        assert uidq.ask() == uid
    # discarding (in asked, and in told)
    uidq.discard("b")
    for uid in ["c", "a", "c", "a"]:
        assert uidq.ask() == uid
    uidq.tell("a")
    uidq.discard("a")
    for uid in ["c", "c"]:
        assert uidq.ask() == uid
    # pickling
    string = pickle.dumps(uidq)
    pickle.loads(string)
    # clearing
    uidq.clear()
    with pytest.raises(RuntimeError):
        uidq.ask()



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import os
import numpy as np
import nevergrad as ng
from . import core
import pytest


@pytest.mark.parametrize("kind", core.OlympusSurface.SURFACE_KINDS)
@pytest.mark.parametrize("noise_kind", ["GaussianNoise", "UniformNoise", "GammaNoise"])
def test_olympus_surface(kind: str, noise_kind: str) -> None:
    try:
        func = core.OlympusSurface(kind=kind, noise_kind=noise_kind)
    except Exception as e:
        if os.name == "nt":
            raise ng.errors.UnsupportedExperiment("Unavailable under Windows.")
        else:
            raise e
    func2 = core.OlympusSurface(kind=kind, noise_kind=noise_kind)  # Let us check the randomization.
    x = 2 * np.random.rand(func.dimension)
    value = func(x)  # should not touch boundaries, so value should be < np.inf
    value2 = func2(x)  # should not touch boundaries, so value should be < np.inf
    assert isinstance(value, float)
    assert value < np.inf
    assert value != value2 or noise_kind == "GammaNoise"


# @pytest.mark.parametrize("dataset_kind", core.OlympusEmulator.DATASETS)
# @pytest.mark.parametrize("model_kind", ["NeuralNet"])  # ["BayesNeuralNet", "NeuralNet"])
# def test_olympus_emulator(dataset_kind: str, model_kind: str) -> None:
#    func = core.OlympusEmulator(dataset_kind=dataset_kind, model_kind=model_kind)
#    x = 2 * np.random.rand(func.dimension)
#    value = func(x)  # should not touch boundaries, so value should be < np.inf
#    assert isinstance(value, float)
#    assert value < np.inf



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
"""Experimental mutation patterns for structured data
"""

import typing as tp
import numpy as np
from . import core
from . import transforms
from .data import Data, Scalar
from .choice import Choice
from . import _layering


D = tp.TypeVar("D", bound=Data)
P = tp.TypeVar("P", bound=core.Parameter)


class Mutation(_layering.Layered):
    """Custom mutation or recombination operation
    This is an experimental API

    Call on a Parameter to create a new Parameter with the
    provided mutation/recombination.
    """

    _TYPE = core.Parameter

    def root(self) -> core.Parameter:
        param = self._layers[0]
        self._check_type(param)
        return param  # type: ignore

    def _check_type(self, param: core.Layered) -> None:
        if not isinstance(param, self._TYPE):
            raise RuntimeError(
                f"{self.__class__.__name__} must be applied to {self._TYPE} parameters, got: {type(param)}"
            )

    def __call__(self, parameter: P, inplace: bool = False) -> P:
        self._check_type(parameter)
        new = parameter if inplace else parameter.copy()
        new.add_layer(self.copy())
        return new


class DataMutation(Mutation):
    _TYPE = Data

    def __init__(self, **parameters: tp.Any) -> None:
        self._parameters = parameters
        super().__init__()

    def root(self) -> Data:  # pylint: disable=useless-super-delegation
        # simpler for typing
        return super().root()  # type: ignore

    def _on_layer_added(self) -> None:
        params = self.root().parameters
        for name, obj in self._parameters.items():
            if name not in params:
                params[name] = core.as_parameter(obj)
        self._parameters = {}


class MutationChoice(DataMutation):
    """Selects one of the provided mutation based on a Choice subparameter
    Caution: there may be subparameter collisions
    """

    def __init__(self, mutations: tp.Sequence[Mutation], with_default: bool = True) -> None:
        self.mutations = list(mutations)
        self.with_default = with_default
        options = [lay.uid for lay in self.mutations] + (["#ROOT#"] if with_default else [])
        super().__init__(mutation_choice=Choice(options))

    def _on_layer_added(self) -> None:
        root = self.root()
        for mut in self.mutations:
            root.add_layer(mut)
        self.mutations = []
        super()._on_layer_added()

    def _select(self) -> core.Layered:
        root = self.root()
        layers = {lay.uid: lay for lay in self._layers}
        layers["#ROOT#"] = root
        return layers[root.parameters["mutation_choice"].value]

    def _layered_recombine(self, *others: core.Layered) -> None:
        self._select()._layered_recombine(*others)

    def _layered_mutate(self) -> None:
        self._select()._layered_mutate()


class Cauchy(Mutation):
    def _layered_mutate(self) -> None:
        root = self.root()
        root.set_standardized_data(root.random_state.standard_cauchy(size=root.dimension))


class Crossover(DataMutation):
    """Operator for merging part of an array into another one

    Parameters
    ----------
    axis: None or int or tuple of ints
        the axis (or axes) on which the merge will happen. This axis will be split into 3 parts: the first and last one will take
        value from the first array, the middle one from the second array.
    max_size: None or int
        maximum size of the part taken from the second array. By default, this is at most around half the number of total elements of the
        array to the power of 1/number of axis.


    Notes
    -----
    - this is experimental, the API may evolve
    - when using several axis, the size of the second array part is the same on each axis (aka a square in 2D, a cube in 3D, ...)

    Examples:
    ---------
    - 2-dimensional array, with crossover on dimension 1:
      0 1 0 0
      0 1 0 0
      0 1 0 0
    - 2-dimensional array, with crossover on dimensions 0 and 1:
      0 0 0 0
      0 1 1 0
      0 1 1 0
    """

    def __init__(
        self,
        axis: tp.Any = None,
        max_size: tp.Union[int, Scalar, None] = None,
        fft: bool = False,
    ) -> None:
        if not isinstance(axis, core.Parameter):
            axis = (axis,) if isinstance(axis, int) else tuple(axis) if axis is not None else None
        super().__init__(max_size=max_size, axis=axis, fft=fft)

    @property
    def axis(self) -> tp.Optional[tp.Tuple[int, ...]]:
        return self.root().parameters["axis"].value  # type: ignore

    def _layered_recombine(self, *arrays: Data) -> None:  # type: ignore
        root = self.root()
        new_value = self._apply_array([root.value] + [a.value for a in arrays])
        bounds = root.bounds
        if root.parameters["fft"].value and any(x is not None for x in bounds):
            new_value = transforms.Clipping(a_min=bounds[0], a_max=bounds[1]).forward(new_value)
        root.value = new_value

    def _apply_array(self, arrays: tp.Sequence[np.ndarray]) -> np.ndarray:
        root = self.root()
        # checks
        if len(arrays) != 2:
            raise Exception(f"Crossover can only be applied between 2 individuals, got {len(arrays)}")
        transf = (
            transforms.Fourrier(range(arrays[0].ndim) if self.axis is None else self.axis)
            if root.parameters["fft"].value
            else None
        )
        if transf is not None:
            arrays = [transf.forward(a) for a in arrays]
        shape = arrays[0].shape
        assert shape == arrays[1].shape, "Individuals should have the same shape"
        # settings
        axis = tuple(range(len(shape))) if self.axis is None else self.axis
        max_size = root.parameters["max_size"].value
        max_size = int(((arrays[0].size + 1) / 2) ** (1 / len(axis))) if max_size is None else max_size
        max_size = min(max_size, *(shape[a] - 1 for a in axis))
        size = 1 if max_size == 1 else self.random_state.randint(1, max_size)
        # slices
        slices = _make_slices(shape, axis, size, self.random_state)
        result = np.array(arrays[0], copy=True)
        result[tuple(slices)] = arrays[1][tuple(slices)]
        if transf is not None:
            result = transf.backward(result)
        return result


class RavelCrossover(Crossover):  # TODO: can be made for all parameters instead of just arrays
    """Operator for merging part of an array into another one, after raveling

    Parameters
    ----------
    max_size: None or int
        maximum size of the part taken from the second array. By default, this is at most around half the number of total elements of the
        array to the power of 1/number of axis.
    """

    def __init__(
        self,
        max_size: tp.Union[int, Scalar, None] = None,
    ) -> None:
        super().__init__(axis=0, max_size=max_size)

    def _apply_array(self, arrays: tp.Sequence[np.ndarray]) -> np.ndarray:
        shape = arrays[0].shape
        out = super()._apply_array([a.ravel() for a in arrays])
        return out.reshape(shape)


def _make_slices(
    shape: tp.Tuple[int, ...], axes: tp.Tuple[int, ...], size: int, rng: np.random.RandomState
) -> tp.List[slice]:
    slices = []
    for a, s in enumerate(shape):
        if a in axes:
            if s <= 1:
                raise ValueError("Cannot crossover on axis with size 1")
            start = rng.randint(s - size)
            slices.append(slice(start, start + size))
        else:
            slices.append(slice(None))
    return slices


class Translation(DataMutation):
    def __init__(self, axis: tp.Optional[tp.Union[int, tp.Iterable[int]]] = None):
        if not isinstance(axis, core.Parameter):
            axes = (axis,) if isinstance(axis, int) else tuple(axis) if axis is not None else None
        super().__init__(axes=axes)

    @property
    def axes(self) -> tp.Optional[tp.Tuple[int, ...]]:
        return self.root().parameters["axes"].value  # type: ignore

    def _layered_mutate(self) -> None:
        root = self.root()
        root._value = self._apply_array([root._value])

    def _apply_array(self, arrays: tp.Sequence[np.ndarray]) -> np.ndarray:
        assert len(arrays) == 1
        data = arrays[0]
        axes = tuple(range(data.ndim)) if self.axes is None else self.axes
        shifts = [self.random_state.randint(data.shape[a]) for a in axes]
        return np.roll(data, shifts, axis=axes)  # type: ignore


class AxisSlicedArray:
    def __init__(self, array: np.ndarray, axis: int):
        self.array = array
        self.axis = axis

    def __getitem__(self, slice_: slice) -> np.ndarray:
        assert isinstance(slice_, slice)
        slices = tuple(slice_ if a == self.axis else slice(None) for a in range(self.array.ndim))
        return self.array[slices]  # type: ignore


class Jumping(DataMutation):
    """Move a chunk for a position to another in an array"""

    def __init__(self, axis: int, size: int):
        super().__init__(axis=axis, size=size)

    @property
    def axis(self) -> int:
        return self.root().parameters["axis"].value  # type: ignore

    @property
    def size(self) -> int:
        return self.root().parameters["size"].value  # type: ignore

    def _layered_mutate(self) -> None:
        root = self.root()
        root._value = self._apply_array([root._value])

    def _apply_array(self, arrays: tp.Sequence[np.ndarray]) -> np.ndarray:
        assert len(arrays) == 1
        data = arrays[0]
        L = data.shape[self.axis]
        size = self.random_state.randint(1, self.size)
        asdata = AxisSlicedArray(data, self.axis)
        init = self.random_state.randint(L)
        chunck = asdata[init : init + size]
        remain: np.ndarray = np.concatenate([asdata[:init], asdata[init + size :]], axis=self.axis)
        # pylint: disable=unsubscriptable-object
        newpos = self.random_state.randint(remain.shape[self.axis])
        asremain = AxisSlicedArray(remain, self.axis)
        return np.concatenate([asremain[:newpos], chunck, asremain[newpos:]], axis=self.axis)  # type: ignore


class LocalGaussian(DataMutation):
    def __init__(
        self, size: tp.Union[int, core.Parameter], axes: tp.Optional[tp.Union[int, tp.Iterable[int]]] = None
    ):
        if not isinstance(axes, core.Parameter):
            axes = (axes,) if isinstance(axes, int) else tuple(axes) if axes is not None else None
        super().__init__(axes=axes, size=size)

    @property
    def axes(self) -> tp.Optional[tp.Tuple[int, ...]]:
        return self.root().parameters["axes"].value  # type: ignore

    def _layered_mutate(self) -> None:
        root = self.root()
        data = np.zeros(root.value.shape)
        # settings
        axis = tuple(range(len(data.shape))) if self.axes is None else self.axes
        size = self.root().parameters["size"].value
        # slices
        slices = _make_slices(data.shape, axis, size, self.random_state)
        shape = data[tuple(slices)].shape
        data[tuple(slices)] += self.random_state.normal(0, 1, size=shape)
        root._internal_set_standardized_data(data.ravel(), reference=root)


def rolling_mean(vector: np.ndarray, window: int) -> np.ndarray:
    if window >= len(vector):
        return np.sum(vector) * np.ones((len(vector),))  # type: ignore
    if window <= 1:
        return vector
    cumsum: np.ndarray = np.cumsum(np.concatenate(([0], vector, vector[: window - 1])))  # type: ignore
    return cumsum[window:] - cumsum[:-window]  # type: ignore


# class TunedTranslation(Mutation):
#     def __init__(self, axis: int, shape: tp.Sequence[int]):
#         assert isinstance(axis, int)
#         self.shape = tuple(shape)
#         super().__init__(shift=Choice(range(1, shape[axis])))
#         self.axis = axis
#
#     @property
#     def shift(self) -> Choice:
#         return self.root().parameters["shift"]  # type: ignore
#
#     def _apply_array(self, arrays: tp.Sequence[np.ndarray]) -> np.ndarray:
#         assert len(arrays) == 1
#         data = arrays[0]
#         assert data.shape == self.shape
#         shift = self.shift.value
#         # update shift arrray
#         shifts = self.shift.indices._value
#         self.shift.indices._value = np.roll(shifts, shift)  # update probas
#         return np.roll(data, shift, axis=self.axis)  # type: ignore


# class ProbaLocalGaussian(Mutation):
#     def __init__(self, axis: int, shape: tp.Sequence[int]):
#         assert isinstance(axis, int)
#         self.shape = tuple(shape)
#         self.axis = axis
#         super().__init__(
#             positions=Array(shape=(shape[axis],)),
#             ratio=Scalar(init=1, lower=0, upper=1).set_mutation(sigma=0.05),
#         )
#
#     def axes(self) -> tp.Optional[tp.Tuple[int, ...]]:
#         return self.root().parameters["axes"].value  # type: ignore
#
#     def apply(self, arrays: tp.Sequence[Data]) -> None:
#         arrays = list(arrays)
#         assert len(arrays) == 1
#         data = np.zeros(arrays[0].value.shape)
#         # settings
#         length = self.shape[self.axis]
#         size = int(max(1, np.round(length * self.root().parameters["ratio"].value)))
#         # slices
#         e_weights = np.exp(rolling_mean(self.root().parameters["positions"].value, size))
#         probas = e_weights / np.sum(e_weights)
#         index = self.random_state.choice(range(length), p=probas)
#         # update (inefficient)
#         shape = tuple(size if a == self.axis else s for a, s in enumerate(arrays[0].value.shape))
#         data[tuple(slice(s) for s in shape)] += self.random_state.normal(0, 1, size=shape)
#         data = np.roll(data, shift=index, axis=self.axis)
#         arrays[0]._internal_set_standardized_data(data.ravel(), reference=arrays[0])



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import typing as tp
import numpy as np
import gym
from . import base


class JamesBond:
    """Holds the state of a player of DoubleOSeven"""

    actions = ["fire", "protect", "reload"]
    max_consecutive_protect = 4

    def __init__(self) -> None:
        self.ammunitions = 0
        self.consecutive_protect = 0

    def update_with_action(self, action: str) -> None:
        """Update the current state with an authorized action."""
        if action not in self.actions:
            raise ValueError(f"Unauthorized action {action}, authorized are {self.actions}")
        if action == "fire":
            if not self.ammunitions:
                raise ValueError("Cannot fire without amunitions")
            self.ammunitions -= 1
        elif action == "reload":
            self.ammunitions += 1
        if action == "protect":
            self.consecutive_protect += 1
        else:
            self.consecutive_protect = 0

    def get_state(self) -> tp.Tuple[int, int]:
        return (self.ammunitions, self.consecutive_protect)


class DoubleOSeven(base.MultiAgentEnv):
    """2-player environment with "player_0" and "player_1", playing the 007 game.
    Each player state has:
    - a number of ammunition (init: 0)
    - a number of consecutive protection (init: 0)
    Each player can:
    - reload: to get one more ammunition
    - fire: to try to kill the other player (reloads instead if no ammunition)
    - protect: to avoid the other player's fire

    The winner is either:
    - the first to fire while the other people is reloading (or firing without ammunition)
    - the one that did not protect more that 4 times consecutively if the other one did.
    The game stops if nobody won after 100 iterations.
    """

    observation_space = gym.spaces.Box(low=0, high=float("inf"), shape=(4,), dtype=int)
    action_space = gym.spaces.Discrete(3)
    agent_names = ["player_0", "player_1"]

    def __init__(self, verbose: bool = False) -> None:
        super().__init__()
        self.players = [JamesBond(), JamesBond()]
        self.verbose = verbose
        self._step = 0

    def reset(self) -> tp.Dict[str, np.ndarray]:
        """Resets the env and returns observations from ready agents.
        Returns:
            obs (dict): New observations for each ready agent.
        """
        self._step = 0
        self.players = [JamesBond(), JamesBond()]
        return self._make_observations()

    def _make_observations(self) -> tp.Dict[str, np.ndarray]:
        states = [p.get_state() for p in self.players]
        return {"player_0": np.array(states[0] + states[1]), "player_1": np.array(states[1] + states[0])}

    def copy(self) -> "DoubleOSeven":
        return self.__class__(verbose=self.verbose)

    def step(self, action_dict: tp.Dict[str, int]) -> base.StepReturn:
        """Returns observations from ready agents.
        The returns are dicts mapping from agent_id strings to values. The
        number of agents in the env can vary over time.
        Returns
        -------
            obs (dict): New observations for each ready agent.
            rewards (dict): Reward values for each ready agent. If the
                episode is just started, the value will be None.
            dones (dict): Done values for each ready agent. The special key
                "__all__" (required) is used to indicate env termination.
            infos (dict): Optional info values for each agent id.
        """
        if self.verbose:
            strings: tp.List[str] = []
            for k in range(2):
                action = JamesBond.actions[action_dict[f"player_{k}"]]
                strings.append(f"Player {k} {self.players[k].get_state()}: {action}")
            print(" - ".join(strings))
        actions = [JamesBond.actions[action_dict[f"player_{k}"]] for k in range(2)]
        self._step += 1
        info: tp.Dict[tp.Any, tp.Any] = {}
        rew = {"player_0": 0, "player_1": 0}
        # change impossible actions
        actions = [
            "reload" if a == "fire" and not p.ammunitions else a for p, a in zip(self.players, actions)
        ]
        # update players
        for player, action in zip(self.players, actions):
            player.update_with_action(action)
        # main way to win
        if actions[0] == "fire" and actions[1] == "reload":
            rew = {"player_0": 1, "player_1": 0}
        elif actions[0] == "reload" and actions[1] == "fire":
            rew = {"player_0": 0, "player_1": 1}
        # lose if you keep protecting
        if any(p.consecutive_protect > JamesBond.max_consecutive_protect for p in self.players):
            if self.players[0].consecutive_protect > self.players[1].consecutive_protect:
                rew = {"player_0": 0, "player_1": 1}
            elif self.players[1].consecutive_protect > self.players[0].consecutive_protect:
                rew = {"player_0": 1, "player_1": 0}
            # if both keep protecting... well, it goes on...
        obs = self._make_observations()
        done = {"__all__": self._step == 100 or sum(abs(x) for x in rew.values()) > 0}
        return obs, rew, done, info



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import pytest
import numpy as np
import sys
from scipy import stats
import nevergrad as ng
import nevergrad.common.typing as tp


# decorators to be used when testing on Windows is unecessary
# or cumbersome
skip_win_perf = pytest.mark.skipif(
    sys.platform == "win32", reason="Slow, and no need to test performance on all platforms"
)


@skip_win_perf  # type: ignore
def test_tabu() -> None:

    num_tests = 97
    for o in ["DiscreteOnePlusOne"]:
        values = []
        valuesT = []
        for _ in range(num_tests):
            dim = 4
            arity = 7
            budget = (arity**dim) // 50
            domain = ng.p.TransitionChoice(range(arity), ordered=False, repetitions=dim)
            optimum = np.random.randint(arity, size=dim)

            def of(x):
                return -np.sum((np.asarray(x) == optimum))

            recom = ng.optimizers.registry[o](domain, budget).minimize(of).value
            recomT = ng.optimizers.registry[o + "T"](domain, budget).minimize(of).value
            values += [of(recom)]
            valuesT += [of(recomT)]
        pval = stats.mannwhitneyu(valuesT, values, alternative="less").pvalue
        assert pval < 0.15, f"{o} fails the Tabu search test: pval = {pval}."


def summation(x: tp.ArrayLike) -> float:
    return sum(x)


@skip_win_perf  # type: ignore
def test_tabu_sum() -> None:

    num_tests = 147
    for o in ["DiscreteOnePlusOne"]:
        values = []
        valuesT = []
        for _ in range(num_tests):
            dim = 24
            arity = 3
            budget = 7
            domain = ng.p.TransitionChoice(range(arity), ordered=False, repetitions=dim)
            domain.tabu_congruence = summation
            optimum = np.random.randint(arity, size=dim)

            def of(x):
                return np.abs(sum(x) - sum(optimum))

            recom = ng.optimizers.registry[o](domain, budget).minimize(of).value
            recomT = ng.optimizers.registry[o + "T"](domain, budget).minimize(of).value
            values += [of(recom)]
            valuesT += [of(recomT)]
        pval = stats.mannwhitneyu(valuesT, values, alternative="less").pvalue
        assert pval < 0.1, f"{o} fails the Tabu search test: pval = {pval}."



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import gym
import numpy as np


class GenericMujocoEnv:
    """This class evaluates policy of OpenAI Gym environment.

    Parameters
    -----------
    env_name: str
        Gym environment name
    state_mean: list
        Average state values of multiple independent runs.
    state_std: list
        Standard deviation of state values of multiple independent runs.
    num_rollouts: int
        number of independent runs.
    activation: str:
        activation function
    layer_rescaling_coef: float
        Scaling coefficient of output layers
    random_state: int or None
        random state for reproducibility in Gym environment.
    """

    def __init__(
        self,
        env_name,
        state_mean,
        state_std,
        num_rollouts,
        activation,
        layer_rescaling_coef,
        noise_level,
        random_state,
    ):
        self.mean = state_mean
        self.std = state_std
        self.env = gym.make(env_name)
        self.num_rollouts = num_rollouts
        self.random_state = random_state
        self.activation = activation
        self.layer_rescaling_coef = layer_rescaling_coef
        self.noise_level = noise_level

    def _activation(self, x):
        if self.activation == "tanh":
            return np.tanh(x)
        elif self.activation == "sigmoid":
            return 1.0 / (1 + np.exp(-x))
        else:
            raise NotImplementedError(r"Activation {self.activation} not implemented.")

    def __call__(self, layers):
        """Compute loss (average cumulative negative reward) of a given policy."""
        returns = []
        for _ in range(self.num_rollouts):
            obs = self.env.reset()
            done = False
            totalr = 0.0
            while not done:
                action = (
                    np.matmul(obs, layers[0])
                    if (self.mean is None)
                    else (np.matmul((obs - self.mean) / self.std, layers[0]))
                )
                action = action * self.layer_rescaling_coef[0]
                for x, r_coef in zip(layers[1:], self.layer_rescaling_coef[1:]):
                    action = np.matmul(self._activation(action) + 1.0e-3, x) * r_coef
                if self.noise_level > 0.0:
                    action += action * self.noise_level * self.random_state.normal(size=action.shape)
                obs, r, done, _ = self.env.step(action)
                totalr += r
            returns.append(totalr)

        return -np.mean(returns)



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import os
import re
import sys
import time
import random
import inspect
import logging
import platform
import tempfile
import warnings
from pathlib import Path
from functools import partial
from unittest import SkipTest
from unittest.mock import patch
import pytest
import numpy as np
import pandas as pd
from scipy import stats
from scipy.ndimage import gaussian_filter
from bayes_opt.util import acq_max

# from bayes_opt.util import NotUniqueError
import nevergrad as ng
import nevergrad.common.typing as tp
from nevergrad.common import testing
from nevergrad.common import errors
from . import base
from . import optimizerlib as optlib
from . import experimentalvariants as xpvariants
from . import es
from .optimizerlib import registry
from .optimizerlib import NGOptBase


# decorators to be used when testing on Windows is unecessary
# or cumbersome
skip_win_perf = pytest.mark.skipif(
    sys.platform == "win32", reason="Slow, and no need to test performance on all platforms"
)


class Fitness:
    """Simple quadratic fitness function which can be used with dimension up to 4"""

    def __init__(self, x0: tp.ArrayLike) -> None:
        self.x0 = np.array(x0, copy=True)
        self.call_times: tp.List[float] = []

    def __call__(self, x: tp.ArrayLike) -> float:
        assert len(self.x0) == len(x)
        self.call_times.append(time.time())
        return float(np.sum((np.array(x, copy=False) - self.x0) ** 2))

    def get_factors(self) -> tp.Tuple[float, float]:
        logdiffs = np.log(np.maximum(1e-15, np.cumsum(np.diff(self.call_times))))
        nums = np.arange(len(logdiffs))
        slope, intercept = (float(np.exp(x)) for x in stats.linregress(nums, logdiffs)[:2])
        return slope, intercept


# pylint: disable=too-many-locals
def check_optimizer(
    optimizer_cls: tp.Union[base.ConfiguredOptimizer, tp.Type[base.Optimizer]],
    budget: int = 300,
    verify_value: bool = True,
) -> None:
    # recast optimizer do not support num_workers > 1, and respect no_parallelization.
    num_workers = 1 if optimizer_cls.recast or optimizer_cls.no_parallelization else 2
    num_attempts = 1 if not verify_value else 3  # allow 3 attemps to get to the optimum (shit happens...)
    optimum = [0.5, -0.8]
    fitness = Fitness(optimum)
    for k in range(1, num_attempts + 1):
        fitness = Fitness(optimum)
        optimizer = optimizer_cls(parametrization=len(optimum), budget=budget, num_workers=num_workers)
        assert isinstance(
            optimizer.provide_recommendation(), ng.p.Parameter
        ), "Recommendation should be available from start"
        with testing.suppress_nevergrad_warnings():
            candidate = optimizer.minimize(fitness)
        raised = False
        if verify_value:
            try:
                np.testing.assert_array_almost_equal(candidate.args[0], optimum, decimal=1)
            except AssertionError as e:
                raised = True
                print(f"Attemp #{k}: failed with best point {tuple(candidate.args[0])}")
                if k == num_attempts:
                    raise e
        if not raised:
            break
    if budget > 100:
        slope, intercept = fitness.get_factors()
        print(f"For your information: slope={slope} and intercept={intercept}")
    # make sure we are correctly tracking the best values
    archive = optimizer.archive
    assert optimizer.current_bests["pessimistic"].pessimistic_confidence_bound == min(
        v.pessimistic_confidence_bound for v in archive.values()
    )
    # add a random point to test tell_not_asked
    assert not optimizer._asked, "All `ask`s  should have been followed by a `tell`"
    try:
        data = np.random.normal(0, 1, size=optimizer.dimension)
        candidate = optimizer.parametrization.spawn_child().set_standardized_data(data)
        optimizer.tell(candidate, 12.0)
    except Exception as e:  # pylint: disable=broad-except
        if not isinstance(e, base.errors.TellNotAskedNotSupportedError):
            raise AssertionError(
                "Optimizers should raise base.TellNotAskedNotSupportedError "
                "at when telling unasked points if they do not support it"
            ) from e
    else:
        assert optimizer.num_tell == budget + 1
        assert optimizer.num_tell_not_asked == 1 or "Smooth" in str(optimizer_cls)


SLOW = [
    "NoisyDE",
    "NoisyBandit",
    "Noisy13Splits",
    "NoisyInfSplits",
    "DiscreteNoisy13Splits",
    "DiscreteNoisyInfSplits",
    "SPSA",
    "NoisyOnePlusOne",
    "OptimisticNoisyOnePlusOne",
    "ASCMADEthird",
    "ASCMA2PDEthird",
    "MultiScaleCMA",
    "PCEDA",
    "EDA",
    "MicroCMA",
    "ES",
]


UNSEEDABLE: tp.List[str] = [
    "CmaFmin2",
    "MetaModelFmin2",
    "NLOPT_GN_CRS2_LM",
    "NLOPT_GN_ISRES",
    "NLOPT_GN_ESCH",
    "GOMEABlock",
    "GOMEA",
    "GOMEATree",
]


def buggy_function(x: np.ndarray) -> float:
    if any(x[::2] > 0.0):
        return float("nan")
    if any(x > 0.0):
        return float("inf")
    return np.sum(x**2)


@pytest.mark.parametrize("dim", [2, 10, 20, 40, 80, 160, 320, 640, 1280, 25600, 51200, 102400])  # type: ignore
@pytest.mark.parametrize("budget_multiplier", [10, 100, 1000, 10000])  # type: ignore
@pytest.mark.parametrize("num_workers", [1, 2, 20])  # type: ignore
@pytest.mark.parametrize("bounded", [False, True])  # type: ignore
@pytest.mark.parametrize("discrete", [False, True])  # type: ignore
def test_ngopt(dim: int, budget_multiplier: int, num_workers: int, bounded: bool, discrete: bool) -> None:
    instrumentation = ng.p.Array(shape=(dim,))
    if bounded:
        instrumentation.set_bounds(lower=-12.0, upper=15.0)
    if discrete:
        instrumentation.set_integer_casting()
    ngopt = optlib.NGOpt(ng.p.Array(shape=(dim,)), budget=budget_multiplier * dim, num_workers=num_workers)
    ngopt.tell(ngopt.ask(), 42.0)


@skip_win_perf  # type: ignore
@pytest.mark.parametrize("name", registry)  # type: ignore
@testing.suppress_nevergrad_warnings()  # hides bad loss
def test_infnan(name: str) -> None:
    if any(x in name for x in ["SMAC", "BO", "AX"]) and os.environ.get("CIRCLECI", False):
        raise SkipTest("too slow for CircleCI!")

    def doint(s):  # Converting a string into an int.
        return 7 + sum([ord(c) * i for i, c in enumerate(s)])

    if doint(name) % 5 > 0:
        raise SkipTest("too many tests for CircleCI!")

    optim_cls = registry[name]
    optim = optim_cls(parametrization=2, budget=70)
    if not (
        any(
            x in name
            for x in [
                "EDA",
                "EMNA",
                "Stupid",
                "NEWUOA",
                "Large",
                "Fmin2",
                "NLOPT",
                "TBPSA",
                "SMAC",
                "BO",
                "Noisy",
                "Chain",
                "chain",  # TODO: remove when possible
            ]
        )
    ):
        recom = optim.minimize(buggy_function)
        result = buggy_function(recom.value)
        if result < 2.0:
            return
        assert (  # The "bad" algorithms, most of them originating in CMA's recommendation rule.
            any(x == name for x in ["WidePSO", "SPSA", "NGOptBase", "Shiwa", "NGO"])
            or isinstance(optim, (optlib.Portfolio, optlib._CMA, optlib.recaster.SequentialRecastOptimizer))
            or "NGOpt" in name
            or "HS" in name
            or "Adapti" in name
            or "MetaModelDiagonalCMA" in name
        )


@skip_win_perf  # type: ignore
@pytest.mark.parametrize("name", registry)  # type: ignore
def test_optimizers(name: str) -> None:
    """Checks that each optimizer is able to converge on a simple test case"""
    if any(x in name for x in ["Chain", "SMAC", "BO", "AX"]) and os.environ.get("CIRCLECI", False):
        raise SkipTest("too slow for CircleCI!")

    def doint(s):  # Converting a string into an int.
        return 7 + sum([ord(c) * i for i, c in enumerate(s)])

    if doint(name) % 5 > 0:
        raise SkipTest("too many tests for CircleCI!")
    if (
        sum([ord(c) for c in name]) % 4 > 0
        and name
        not in [
            "DE",
            "CMA",
            "OnePlusOne",
            "Cobyla",
            "DiscreteLenglerOnePlusOne",
            "PSO",
        ]
        or "Tiny" in name
        or "Micro" in name
    ) and os.environ.get("CIRCLECI", False):
        raise SkipTest("Too expensive: we randomly skip 3/4 of these tests.")
    if name in ["CMAbounded", "NEWUOA"]:  # Not a general purpose optimization method.
        return
    if "BO" in name:  # Bayesian Optimization is rarely good, let us save up time.
        return
    optimizer_cls = registry[name]
    if isinstance(optimizer_cls, base.ConfiguredOptimizer):
        assert any(
            hasattr(mod, name) for mod in (optlib, xpvariants)
        )  # make sure registration matches name in optlib/xpvariants
        assert (
            optimizer_cls.__class__(**optimizer_cls._config) == optimizer_cls
        ), "Similar configuration are not equal"
    # some classes of optimizer are eigher slow or not good with small budgets:
    nameparts = ["Many", "Chain", "BO", "Discrete", "NLOPT"] + ["chain"]  # TODO remove chain when possible
    is_ngopt = inspect.isclass(optimizer_cls) and issubclass(optimizer_cls, NGOptBase)  # type: ignore
    verify = (
        not optimizer_cls.one_shot
        and name not in SLOW
        and not any(x in name for x in nameparts)
        and not is_ngopt
    )
    budget = 300 if "BO" not in name and not is_ngopt else 4
    # the following context manager speeds up BO tests
    patched = partial(acq_max, n_warmup=10000, n_iter=2)
    with patch("bayes_opt.bayesian_optimization.acq_max", patched):
        check_optimizer(optimizer_cls, budget=budget, verify_value=verify)


@pytest.mark.parametrize("name", registry)  # type: ignore
def test_optimizers_minimal(name: str) -> None:
    optimizer_cls = registry[name]
    if any(x in name for x in ["SMAC", "BO", "AX"]) and os.environ.get("CIRCLECI", False):
        raise SkipTest("too slow for CircleCI!")
    if optimizer_cls.one_shot or name in ["CM", "NLOPT_LN_PRAXIS", "ES", "RecMixES", "RecMutDE", "RecES"]:
        return
    if any(
        x in str(optimizer_cls)
        for x in [
            "BO",
            "BAR",
            "Meta",
            "Voronoi",
            "tuning",
            "QrDE",
            "BIPOP",
            "ECMA",
            "CMAstd",
            "_COBYLA",
            "HyperOpt",
            "Chain",
            "CMAbounded",
            "Tiny",
            "iscrete",
            "GOMEA",
            "para",
            "SPSA",
            "EDA",
            "FCMA",
            "Noisy",
            "HS",
            "SQPCMA",
            "CMandAS",
            "GA",
            "EMNA",
            "RL",
            "Milli",
            "Small",
            "small",
            "Chain",
            "Tree",
            "Mix",
            "Micro",
            "Naive",
            "Portfo",
            "ESCH",
            "Multi",
            "NGO",
            "Discrete",
            "MixDet",
            "Rotated",
            "Iso",
            "Bandit",
            "TBPSA",
            "VLP",
            "LPC",
            "Choice",
            "NLOPT_GN_ISRES",
        ]
    ):
        raise SkipTest("Skipped for saving up CircleCI resources!")

    def f(x):
        return sum((x - 1.1) ** 2)

    def mf(x):
        return sum((x + 1.1) ** 2)

    def f1(x):
        return (x - 1.1) ** 2

    def f2(x):
        return (x + 1.1) ** 2

    def f1p(x):
        return sum((x - 1.1) ** 2)

    def f2p(x):
        return sum((x + 1.1) ** 2)

    if (
        "BAR" in name or "Cma" in name or "CMA" in name or "BIPOP" in name
    ):  # Sometimes CMA does not work in dim 1 :-(
        budget = 600
        if "BAR" in name:
            budget = 3600
        if any(x in name for x in ["Large", "Tiny", "Para", "Diagonal"]):
            return
        val = optimizer_cls(2, budget).minimize(f).value[0]
        assert (1.04 < val < 1.16) or (val > 1.0 and "BAR" in name), f"pb with {optimizer_cls} for 1.1: {val}"
        val = optimizer_cls(2, budget).minimize(mf).value[0]
        assert -1.17 < val < -1.04 or (
            val < -1.04 and "BAR" in name
        ), f"pb with {optimizer_cls} for -1.1.: {val}"
        v = ng.p.Array(shape=(2,), upper=1.0, lower=0.0)
        val = optimizer_cls(v, budget).minimize(f1p).value[0]
        assert 0.90 < val < 1.01, f"pb with {optimizer_cls} for 1.: {val}"
        v = ng.p.Array(shape=(2,), upper=0.3, lower=-0.3)
        val = optimizer_cls(v, budget).minimize(f2p).value[0]
        assert -0.31 < val < -0.24, f"pb with {optimizer_cls} for -0.3: {val}"
    else:
        budget = 100
        if "DE" in name or "PSO" in name or "Hyper" in name:
            budget = 300
        if any(x in name for x in ["QO", "SODE"]):
            return
        val = optimizer_cls(1, budget).minimize(f).value
        assert (1.04 < val < 1.16) or (val > 1.0 and "BAR" in name), f"pb with {optimizer_cls} for 1.1: {val}"
        val = optimizer_cls(1, budget).minimize(mf).value
        assert -1.16 < val < -1.04, f"pb with {optimizer_cls} for -1.1.: {val}"
        v = ng.p.Scalar(upper=1.0, lower=0.0)  # type: ignore
        val = optimizer_cls(v, budget).minimize(f1).value
        assert 0.94 < val < 1.01, f"pb with {optimizer_cls} for 1.: {val}"
        v = ng.p.Scalar(upper=0.3, lower=-0.3)  # type: ignore
        val = optimizer_cls(v, budget).minimize(f2).value
        assert -0.31 < val < -0.24, f"pb with {optimizer_cls} for -0.3: {val}"


class RecommendationKeeper:
    def __init__(self, filepath: Path) -> None:
        self.filepath = filepath
        self.recommendations = pd.DataFrame(columns=[f"v{k}" for k in range(16)])  # up to 64 values
        if filepath.exists():
            self.recommendations = pd.read_csv(filepath, index_col=0)

    def save(self) -> None:
        # sort and remove unused names
        # then update recommendation file
        names = sorted(x for x in self.recommendations.index if x in registry)
        recom = self.recommendations.loc[names, :]
        recom.iloc[:, :] = np.round(recom, 10)
        recom.to_csv(self.filepath)


@pytest.fixture(scope="module")  # type: ignore
def recomkeeper() -> tp.Generator[RecommendationKeeper, None, None]:
    keeper = RecommendationKeeper(filepath=Path(__file__).parent / "recorded_recommendations.csv")
    yield keeper
    keeper.save()


# pylint: disable=redefined-outer-name
@pytest.mark.parametrize("name", registry)  # type: ignore
def test_optimizers_recommendation(name: str, recomkeeper: RecommendationKeeper) -> None:
    if any(x in name for x in ["SMAC", "BO", "AX"]) and os.environ.get("CIRCLECI", False):
        raise SkipTest("too slow for CircleCI!")
    if name in UNSEEDABLE:
        raise SkipTest("Not playing nicely with the tests (unseedable)")
    if "BO" in name or "EDA" in name:
        raise SkipTest("BO differs from one computer to another")
    if "SMAC" in name:
        raise SkipTest("SMAC is too slow for the 20s limit")
    if len(name) > 8:
        raise SkipTest("Let us check only compact methods.")
    # set up environment
    optimizer_cls = registry[name]
    np.random.seed(None)
    if optimizer_cls.recast:
        np.random.seed(12)
        random.seed(12)  # may depend on non numpy generator
    # budget=6 by default, larger for special cases needing more
    budget = {"WidePSO": 100, "PSO": 200, "MEDA": 100, "EDA": 100, "MPCEDA": 100, "TBPSA": 100}.get(name, 6)
    if isinstance(optimizer_cls, (optlib.DifferentialEvolution, optlib.EvolutionStrategy)):
        budget = 80
    dimension = min(16, max(4, int(np.sqrt(budget))))
    # set up problem
    fitness = Fitness([0.5, -0.8, 0, 4] + (5 * np.cos(np.arange(dimension - 4))).tolist())
    with testing.suppress_nevergrad_warnings():
        optim = optimizer_cls(parametrization=dimension, budget=budget, num_workers=1)
        optim.parametrization.random_state.seed(12)
        np.testing.assert_equal(optim.name, name)
        # the following context manager speeds up BO tests
        # BEWARE: BO tests are deterministic but can get different results from a computer to another.
        # Reducing the precision could help in this regard.
        # patched = partial(acq_max, n_warmup=10000, n_iter=2)
        # with patch("bayes_opt.bayesian_optimization.acq_max", patched):
        recom = optim.minimize(fitness)
    if name not in recomkeeper.recommendations.index:
        # recomkeeper.recommendations.loc[name, :dimension] = tuple(recom.value)
        for i in range(len(recom.value)):
            recomkeeper.recommendations.loc[name, f"v{i}"] = recom.value[i]
        # cummax.loc[:i, 'atr_Lts'] =
        # cummax.iloc[:i].loc[:, 'atr_Lts'] =
        raise ValueError(
            f'Recorded the value {tuple(recom.value)} for optimizer "{name}", please rerun this test locally.'
        )
    # BO slightly differs from a computer to another
    decimal = 2 if isinstance(optimizer_cls, optlib.ParametrizedBO) or "BO" in name else 5
    np.testing.assert_array_almost_equal(
        recom.value,
        np.array(recomkeeper.recommendations.loc[name, :][:dimension], float),
        decimal=decimal,
        err_msg="Something has changed, if this is normal, delete the following "
        f"file and rerun to update the values:\n{recomkeeper.filepath}",
    )
    # check that by default the recommendation has been evaluated
    if isinstance(optimizer_cls, optlib.EvolutionStrategy):  # no noisy variants
        assert recom.loss is not None


@testing.parametrized(
    de=("DE", 10, 10, 30),
    de_w=("DE", 50, 40, 40),
    de1=("OnePointDE", 10, 10, 30),
    de1_w=("OnePointDE", 50, 40, 40),
    dim_d=("AlmostRotationInvariantDEAndBigPop", 50, 40, 51),
    dim=("AlmostRotationInvariantDEAndBigPop", 10, 40, 40),
    dim_d_rot=("RotationInvariantDE", 50, 40, 51),
    large=("BPRotationInvariantDE", 10, 40, 70),
)
def test_differential_evolution_popsize(name: str, dimension: int, num_workers: int, expected: int) -> None:
    optim = registry[name](parametrization=dimension, budget=100, num_workers=num_workers)
    np.testing.assert_equal(optim.llambda, expected)  # type: ignore


@testing.suppress_nevergrad_warnings()
def test_portfolio_budget() -> None:
    for k in range(3, 13):
        optimizer = optlib.Portfolio(parametrization=2, budget=k)
        np.testing.assert_equal(optimizer.budget, sum(o.budget for o in optimizer.optims))  # type: ignore


def test_optimizer_families_repr() -> None:
    Cls = optlib.DifferentialEvolution
    np.testing.assert_equal(repr(Cls()), "DifferentialEvolution()")
    np.testing.assert_equal(repr(Cls(initialization="LHS")), "DifferentialEvolution(initialization='LHS')")
    #
    optim: base.ConfiguredOptimizer = optlib.RandomSearchMaker(sampler="cauchy")
    np.testing.assert_equal(repr(optim), "RandomSearchMaker(sampler='cauchy')")
    #
    optim = optlib.NonObjectOptimizer(method="COBYLA")
    np.testing.assert_equal(repr(optim), "NonObjectOptimizer(method='COBYLA')")
    assert optim.no_parallelization
    #
    optim = optlib.ParametrizedCMA(diagonal=True)
    np.testing.assert_equal(repr(optim), "ParametrizedCMA(diagonal=True)")
    #
    optim = optlib.NoisySplit(discrete=True)
    np.testing.assert_equal(repr(optim), "NoisySplit(discrete=True)")
    assert optim._OptimizerClass.multivariate_optimizer is optlib.OptimisticDiscreteOnePlusOne  # type: ignore


@pytest.mark.parametrize("name", ["PSO", "DE"])  # type: ignore
def test_tell_not_asked(name: str) -> None:
    param = ng.p.Scalar()
    with testing.suppress_nevergrad_warnings():
        opt = optlib.registry[name](parametrization=param, budget=2, num_workers=2)
    opt.llambda = 2  # type: ignore
    t_10 = opt.parametrization.spawn_child(new_value=10)
    t_100 = opt.parametrization.spawn_child(new_value=100)
    assert not opt.population  # type: ignore
    opt.tell(t_10, 90)  # not asked
    assert len(opt.population) == 1  # type: ignore
    asked = opt.ask()
    opt.tell(asked, 88)
    assert len(opt.population) == 2  # type: ignore
    opt.tell(t_100, 0)  # not asked
    asked = opt.ask()
    opt.tell(asked, 89)
    assert len(opt.population) == 2  # type: ignore
    assert opt.num_tell == 4, opt.num_tell
    assert opt.num_ask == 2
    assert len(opt.population) == 2  # type: ignore
    assert int(opt.recommend().value) == 100
    if isinstance(opt.population, dict):  # type: ignore
        assert t_100.uid in opt.population  # type: ignore
    for point, value in opt.archive.items_as_arrays():
        assert value.count == 1, f"Error for point {point}"


def test_tbpsa_recom_with_update() -> None:
    budget = 20
    # set up problem
    fitness = Fitness([0.5, -0.8, 0, 4])
    optim = optlib.TBPSA(parametrization=4, budget=budget, num_workers=1)
    optim.parametrization.random_state.seed(12)
    optim.popsize.llambda = 3  # type: ignore
    candidate = optim.minimize(fitness)
    np.testing.assert_almost_equal(candidate.args[0], [0.037964, 0.0433031, -0.4688667, 0.3633273])


def _square(x: np.ndarray, y: float = 12) -> float:
    return sum((x - 0.5) ** 2) + abs(y)


def _smooth_target(x: np.ndarray) -> float:
    result = 0.0
    d = len(x)
    for h in range(d):
        for v in range(d):
            val = x[h][v]
            assert np.abs(val) <= 1.0
            target = h / d - v / d
            result += 1.0 if np.abs(target - val) > 0.1 else 0.0
    return result


def test_optimization_doc_parametrization_example() -> None:
    instrum = ng.p.Instrumentation(ng.p.Array(shape=(2,)), y=ng.p.Scalar())
    optimizer = optlib.OnePlusOne(parametrization=instrum, budget=100)
    recom = optimizer.minimize(_square)
    assert len(recom.args) == 1
    testing.assert_set_equal(recom.kwargs, ["y"])
    value = _square(*recom.args, **recom.kwargs)
    assert value < 0.25  # should be large enough by an order of magnitude (but is not :s)


def test_optimization_discrete_with_one_sample() -> None:
    optimizer = optlib.PortfolioDiscreteOnePlusOne(parametrization=1, budget=10)
    optimizer.minimize(_square)


# def test_smooth_discrete_one_plus_one() -> None:
#    n = 35
#    d = 35
#    budget = d * d // 2
#    parametrization = ng.p.Array(shape=(d, d), upper=1.0, lower=-1.0)
#    values = []
#    values_smooth = []
#    for _ in range(n):
#        optimizer = xpvariants.SmoothDiscreteOnePlusOne(parametrization=parametrization, budget=budget)
#        recom_smooth = optimizer.minimize(_smooth_target).value
#        optimizer = optlib.DiscreteOnePlusOne(parametrization=parametrization, budget=budget)
#        recom = optimizer.minimize(_smooth_target).value
#        values_smooth += [_smooth_target(recom_smooth)]
#        values += [_smooth_target(recom)]
#    pval = stats.mannwhitneyu(values_smooth, values, alternative="less").pvalue
#    print(f"pval={pval}")
#    assert pval < 0.4, f"P-Value for smooth methods = {pval}."


@pytest.mark.parametrize("name", ["TBPSA", "PSO", "TwoPointsDE", "CMA", "BO"])  # type: ignore
def test_optim_pickle(name: str) -> None:
    # some generic class can fail to be pickled:
    # example of work around:
    # "self.population = base.utils.Population[DEParticle]([])"
    # becomes:
    # "self.population: base.utils.Population[DEParticle] = base.utils.Population([])""
    #
    # Scipy optimizers also fail to be pickled, but this is more complex to solve (not supported yet)
    optim = registry[name](parametrization=12, budget=100, num_workers=2)
    with tempfile.TemporaryDirectory() as folder:
        optim.dump(Path(folder) / "dump_test.pkl")


def test_bo_parametrization_and_parameters() -> None:
    # parametrization
    parametrization = ng.p.Instrumentation(ng.p.Choice([True, False]))
    with pytest.warns(errors.InefficientSettingsWarning):
        xpvariants.QRBO(parametrization, budget=10)
    with pytest.warns(None) as record:  # type: ignore
        opt = optlib.ParametrizedBO(gp_parameters={"alpha": 1})(parametrization, budget=10)
    assert not record, record.list  # no warning

    # parameters
    # make sure underlying BO optimizer gets instantiated correctly
    new_candidate = opt.parametrization.spawn_child(new_value=((True,), {}))
    opt.tell(new_candidate, 0.0)


def test_bo_init() -> None:
    if platform.system() == "Windows":
        raise SkipTest("This test fails on Windows, no idea why.")
    arg = ng.p.Scalar(init=4, lower=1, upper=10).set_integer_casting()
    # The test was flaky with normalize_y=True.
    gp_param = {"alpha": 1e-5, "normalize_y": False, "n_restarts_optimizer": 1, "random_state": None}
    my_opt = ng.optimizers.ParametrizedBO(gp_parameters=gp_param, initialization=None)
    try:
        optimizer = my_opt(parametrization=arg, budget=10)
        optimizer.minimize(np.abs)
    #    except NotUniqueError:
    except Exception as e:
        print(f"Problem {e} in Bayesian optimization.")  # Anyway Bayesian Optimization is basically weak.


def test_chaining() -> None:
    budgets = [7, 19]
    optimizer = optlib.Chaining([optlib.LHSSearch, optlib.HaltonSearch, optlib.OnePlusOne], budgets)(2, 40)
    optimizer.minimize(_square)
    expected = [(7, 7, 0), (19, 19 + 7, 7), (14, 14 + 19 + 7, 19 + 7)]
    for (ex_ask, ex_tell, ex_tell_not_asked), opt in zip(expected, optimizer.optimizers):  # type: ignore
        assert opt.num_ask == ex_ask
        assert opt.num_tell == ex_tell
        assert opt.num_tell_not_asked == ex_tell_not_asked
    optimizer.ask()
    assert optimizer.optimizers[-1].num_ask == 15  # type: ignore


def test_parametrization_optimizer_reproducibility() -> None:
    parametrization = ng.p.Instrumentation(ng.p.Array(shape=(1,)), y=ng.p.Choice(list(range(100))))
    parametrization.random_state.seed(12)
    optimizer = optlib.RandomSearch(parametrization, budget=20)
    recom = optimizer.minimize(_square)
    np.testing.assert_equal(recom.kwargs["y"], 1)
    # resampling deterministically, to make sure it is identical
    data = recom.get_standardized_data(reference=optimizer.parametrization)
    recom = optimizer.parametrization.spawn_child()
    with ng.p.helpers.deterministic_sampling(recom):
        recom.set_standardized_data(data)
    np.testing.assert_equal(recom.kwargs["y"], 1)


@testing.suppress_nevergrad_warnings()
def test_parallel_es() -> None:
    opt = optlib.EvolutionStrategy(popsize=3, offsprings=None)(4, budget=20, num_workers=5)
    for k in range(35):
        cand = opt.ask()  # asking should adapt to the parallelization
        if not k:
            opt.tell(cand, 1)


class QuadFunction:
    """Quadratic function for testing purposes"""

    def __init__(self, scale: float, ellipse: bool) -> None:
        self.scale = scale
        self.ellipse = ellipse

    def __call__(self, x: np.ndarray) -> float:
        y = x - self.scale
        if self.ellipse:
            y *= np.arange(1, x.size + 1) ** 2
        return float(sum(y**2))


META_TEST_ARGS = "dimension,num_workers,scale,budget,ellipsoid".split(",")


def get_metamodel_test_settings(seq: bool = False, special: bool = False):
    tests_metamodel = [
        (2, 8, 1.0, 120, False),
        (2, 3, 8.0, 130, True),
        (5, 1, 1.0, 150, False),
    ]
    if special:
        # Interesting tests removed from CircleCI for flakiness (and we do stats when not on CircleCI):
        tests_metamodel += [
            (8, 27, 8.0, 380, True),
            (2, 1, 8.0, 120, True),
            (2, 3, 8.0, 70, False),
            (1, 1, 1.0, 20, True),
            (1, 3, 5.0, 20, False),
            (2, 3, 1.0, 70, True),
            (2, 1, 8.0, 40, False),
            (5, 3, 1.0, 225, True),
            (5, 1, 8.0, 150, False),
            (5, 3, 8.0, 500, True),
            (9, 27, 8.0, 700, True),
            (10, 27, 8.0, 400, False),
        ]
    if seq:
        for i, (d, _, s, b, e) in enumerate(tests_metamodel):
            tests_metamodel[i] = (d, 1, s, b, e)
    return tests_metamodel


@testing.suppress_nevergrad_warnings()
@skip_win_perf  # type: ignore
@pytest.mark.parametrize("args", get_metamodel_test_settings())
@pytest.mark.parametrize("baseline", ("CMA", "ECMA"))
def test_metamodel(baseline: str, args: tp.Tuple[tp.Any, ...]) -> None:
    """The test can operate on the sphere or on an elliptic funciton."""
    kwargs = dict(zip(META_TEST_ARGS, args))
    check_metamodel(baseline=baseline, **kwargs)


def check_metamodel(
    dimension: int,
    num_workers: int,
    scale: float,
    budget: int,
    ellipsoid: bool,
    baseline: str,
    num_trials: int = 1,
) -> None:
    """This check is called in parametrized tests, with several different parametrization
    (see test_special.py)
    """
    target = QuadFunction(scale=scale, ellipse=ellipsoid)
    # In both cases we compare MetaModel and CMA for a same given budget.
    # But we expect MetaModel to be clearly better only for a larger budget in the ellipsoid case.
    contextual_budget = budget if ellipsoid else 3 * budget
    contextual_budget *= int(max(1, np.sqrt(scale)))
    successes = 0
    for _ in range(num_trials):
        if successes > num_trials // 2:  # We already have enough
            break

        # Let us run the comparison.
        recommendations: tp.List[np.ndarray] = []
        for name in ("MetaModel", baseline if dimension > 1 else "OnePlusOne"):
            opt = registry[name](dimension, contextual_budget, num_workers=num_workers)
            recommendations.append(opt.minimize(target).value)
        metamodel_recom, default_recom = recommendations  # pylint: disable=unbalanced-tuple-unpacking

        # Let us assert that MetaModel is better.
        if target(default_recom) < target(metamodel_recom):
            continue

        # With large budget, the difference should be significant.
        if budget > 60 * dimension:
            if not target(default_recom) > 4.0 * target(metamodel_recom):
                continue

        # ... even more in the non ellipsoid case.
        if budget > 60 * dimension and not ellipsoid:
            if not target(default_recom) > 7.0 * target(metamodel_recom):
                continue
        successes += 1
        assert successes > num_trials // 2, f"Problem for beating {baseline}."


@pytest.mark.parametrize(  # type: ignore
    "penalization,expected,as_layer",
    [
        (False, [1.005573e00, 3.965783e-04], False),
        (True, [0.999975, -0.111235], False),
        (False, [1.000132, -3.679e-4], True),
    ],
)
@testing.suppress_nevergrad_warnings()  # hides failed constraints
def test_constrained_optimization(penalization: bool, expected: tp.List[float], as_layer: bool) -> None:
    def constraint(i: tp.Any) -> tp.Union[bool, float]:
        if penalization:
            return -float(abs(i[1]["x"][0] - 1))
        out = i[1]["x"][0] >= 1
        return out if not as_layer else float(not out)

    parametrization = ng.p.Instrumentation(x=ng.p.Array(shape=(1,)), y=ng.p.Scalar())
    optimizer = optlib.OnePlusOne(parametrization, budget=100)
    optimizer.parametrization.random_state.seed(12)
    if penalization:
        optimizer._constraints_manager.update(max_trials=10, penalty_factor=10)

    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", category=UserWarning)
        optimizer.parametrization.register_cheap_constraint(constraint, as_layer=as_layer)
    recom = optimizer.minimize(_square, verbosity=2)
    np.testing.assert_array_almost_equal([recom.kwargs["x"][0], recom.kwargs["y"]], expected)


@pytest.mark.parametrize("name", registry)  # type: ignore
def test_parametrization_offset(name: str) -> None:
    if any(x in name for x in ["SMAC", "BO", "AX"]) and os.environ.get(
        "CIRCLECI", False
    ):  # Outside CircleCI, only the big.
        raise SkipTest("too slow for CircleCI!")
    if sum([ord(c) for c in name]) % 4 > 0:
        raise SkipTest("Randomly skipping 75% of these tests.")
    if "PSO" in name or "BO" in name:
        raise SkipTest("PSO and BO have large initial variance")
    if "Cobyla" in name and platform.system() == "Windows":
        raise SkipTest("Cobyla is flaky on Windows for unknown reasons")
    parametrization = ng.p.Instrumentation(ng.p.Array(init=[1e12, 1e12]))
    with testing.suppress_nevergrad_warnings():
        optimizer = registry[name](parametrization, budget=100, num_workers=1)
    if optimizer.parametrization.tabu_length > 0:
        return
    for k in range(10 if "BO" not in name else 2):
        candidate = optimizer.ask()
        assert (
            candidate.args[0][0] > 100
        ), f"Candidate value[0] at iteration #{k} is below 100: {candidate.value}"
        optimizer.tell(candidate, 0)


def test_optimizer_sequence() -> None:
    budget = 24
    parametrization = ng.p.Tuple(*(ng.p.Scalar(lower=-12, upper=12) for _ in range(2)))
    optimizer = optlib.LHSSearch(parametrization, budget=budget)
    points = [np.array(optimizer.ask().value) for _ in range(budget)]
    assert sum(any(abs(x) > (budget // 2) - 1 for x in p) for p in points) > 0


def test_shiwa_dim1() -> None:
    param = ng.p.Log(lower=1, upper=1000).set_integer_casting()
    init = param.value
    optimizer = optlib.Shiwa(param, budget=40)
    recom = optimizer.minimize(np.abs)
    assert recom.value < init


continuous_cases: tp.List[tp.Tuple[str, object, int, int, str]] = [
    ("NGOpt", d, b, n, f"#CONTINUOUS")
    for d in [1, 2, 10, 100, 1000]
    for b in [2 * d, 10 * d, 100 * d]
    for n in [1, d, 10 * d]
]


@pytest.mark.parametrize(  # type: ignore
    "name,param,budget,num_workers,expected",
    [
        ("Shiwa", 1, 10, 1, "Cobyla"),
        ("Shiwa", 1, 10, 2, "OnePlusOne"),
        (
            "Shiwa",
            ng.p.Log(lower=1, upper=1000).set_integer_casting(),
            10,
            2,
            "DoubleFastGADiscreteOnePlusOne",
        ),
        ("NGOpt8", ng.p.TransitionChoice(range(30), repetitions=10), 10, 2, "CMandAS2"),
        ("NGOpt8", ng.p.TransitionChoice(range(3), repetitions=10), 10, 2, "AdaptiveDiscreteOnePlusOne"),
        ("NGO", 1, 10, 1, "Cobyla"),
        ("NGO", 1, 10, 2, "OnePlusOne"),
    ]
    + continuous_cases,  # pylint: disable=too-many-arguments
)
@testing.suppress_nevergrad_warnings()
def test_ngopt_selection(
    name: str, param: tp.Any, budget: int, num_workers: int, expected: str, caplog: tp.Any
) -> None:
    with caplog.at_level(logging.DEBUG, logger="nevergrad.optimization.optimizerlib"):
        # pylint: disable=expression-not-assigned
        opt = optlib.registry[name](param, budget=budget, num_workers=num_workers)
        # pylint: disable=pointless-statement
        opt.optim  # type: ignore
        pattern = rf".*{name} selected (?P<name>\w+?) optimizer\."
        match = re.match(pattern, caplog.text.splitlines()[-1])
        assert match is not None, f"Did not detect selection in logs: {caplog.text}"
        choice = match.group("name")
        if expected != "#CONTINUOUS":
            assert choice == expected
        else:
            print(f"Continuous param={param} budget={budget} workers={num_workers} --> {choice}")
            if num_workers >= budget > 600:
                assert choice == "MetaTuneRecentering"
            if num_workers > 1:
                assert choice not in ["SQP", "Cobyla"]
        if "CMA" not in choice:
            assert choice == opt._info()["sub-optim"]
        else:
            assert choice in opt._info()["sub-optim"]


def test_bo_ordering() -> None:
    with testing.suppress_nevergrad_warnings():  # tests do not need to be efficient
        optim = ng.optimizers.ParametrizedBO(initialization="Hammersley")(
            parametrization=ng.p.Choice(range(12)), budget=10
        )
    cand = optim.ask()
    optim.tell(cand, 12)
    optim.provide_recommendation()


@skip_win_perf  # type: ignore
@pytest.mark.parametrize(  # type: ignore
    "name,dimension,num_workers,fake_learning,budget,expected",
    [
        ("NGOpt8", 3, 1, False, 100, ["OnePlusOne", "OnePlusOne"]),
        ("NGOpt8", 3, 1, False, 200, ["SQP", "SQP"]),
        ("NGOpt8", 3, 1, True, 1000, ["SQP", "monovariate", "monovariate"]),
        (None, 3, 1, False, 1000, ["CMA", "OnePlusOne"]),
        (None, 3, 20, False, 1000, ["MetaModel", "OnePlusOne"]),
    ],
)
def test_ngo_split_optimizer(
    name: tp.Optional[str],
    dimension: int,
    num_workers: int,
    fake_learning: bool,
    budget: int,
    expected: tp.List[str],
) -> None:
    if fake_learning:
        param: ng.p.Parameter = ng.p.Instrumentation(
            # a log-distributed scalar between 0.001 and 1.0
            learning_rate=ng.p.Log(lower=0.001, upper=1.0),
            # an integer from 1 to 12
            batch_size=ng.p.Scalar(lower=1, upper=12).set_integer_casting(),
            # either "conv" or "fc"
            architecture=ng.p.Choice(["conv", "fc"]),
        )
    else:
        param = ng.p.Choice(["const", ng.p.Array(init=list(range(dimension)))])
    opt: base.OptCls = (
        xpvariants.MetaNGOpt10
        if name is None
        else (optlib.ConfSplitOptimizer(multivariate_optimizer=optlib.registry[name]))
    )
    optimizer = opt(param, budget=budget, num_workers=num_workers)
    expected = [x if x != "monovariate" else optimizer._config.monovariate_optimizer.name for x in expected]  # type: ignore
    assert optimizer._info()["sub-optim"] == ",".join(expected) or ("CMA" in optimizer._info()["sub-optim"])


@skip_win_perf  # type: ignore
@pytest.mark.parametrize(  # type: ignore
    "budget,with_int",
    [
        (150, True),
        (200, True),
        (666, True),
        (2000, True),
        (66, False),
        (200, False),
        (666, False),
        (2000, False),
    ],
)
def test_ngopt_on_simple_realistic_scenario(budget: int, with_int: bool) -> None:
    if sum([ord(c) for c in f"{budget}-{with_int}"]) % 4 > 0:
        raise SkipTest("Randomly skipping 75% of these tests.")

    def fake_training(learning_rate: float, batch_size: int, architecture: str) -> float:
        # optimal for learning_rate=0.2, batch_size=4, architecture="conv"
        return (learning_rate - 0.2) ** 2 + (batch_size - 4) ** 2 + (0 if architecture == "conv" else 10)

    # Instrumentation class is used for functions with multiple inputs
    # (positional and/or keywords)
    parametrization = ng.p.Instrumentation(
        # a log-distributed scalar between 0.001 and 1.0
        learning_rate=ng.p.Log(lower=0.001, upper=1.0),
        # an integer from 1 to 12
        batch_size=ng.p.Scalar(lower=1, upper=12).set_integer_casting()
        if with_int
        else ng.p.Scalar(lower=1, upper=12),
        # either "conv" or "fc"
        architecture=ng.p.Choice(["conv", "fc"]),
    )

    optimizer = ng.optimizers.NGOpt(parametrization=parametrization, budget=budget)
    recommendation = optimizer.minimize(fake_training)
    result = fake_training(**recommendation.kwargs)
    assert result < 1.0006 if with_int else 5e-3, f"{result} not < {1.0006 if with_int else 5e-3}"


def _multiobjective(z: np.ndarray) -> tp.Tuple[float, float, float]:
    x, y = z
    return (abs(x - 1), abs(y + 1), abs(x - y))


@pytest.mark.parametrize("name", ["DE", "ES", "OnePlusOne"])  # type: ignore
@testing.suppress_nevergrad_warnings()  # hides bad loss
def test_mo_constrained(name: str) -> None:
    optimizer = optlib.registry[name](2, budget=60)
    optimizer.parametrization.random_state.seed(12)

    def constraint(arg: tp.Any) -> bool:  # pylint: disable=unused-argument
        """Random constraint to mess up with the optimizer"""
        return bool(optimizer.parametrization.random_state.rand() > 0.8)

    optimizer.parametrization.register_cheap_constraint(constraint)
    optimizer.minimize(_multiobjective)
    point = optimizer.parametrization.spawn_child(new_value=np.array([1.0, 1.0]))  # on the pareto
    optimizer.tell(point, _multiobjective(point.value))
    if isinstance(optimizer, es._EvolutionStrategy):
        assert optimizer._rank_method is not None  # make sure the nsga2 ranker is used


@pytest.mark.parametrize("name", ["DE", "ES", "OnePlusOne"])  # type: ignore
@testing.suppress_nevergrad_warnings()  # hides bad loss
def test_mo_with_nan(name: str) -> None:
    param = ng.p.Instrumentation(x=ng.p.Scalar(lower=0, upper=5), y=ng.p.Scalar(lower=0, upper=3))
    optimizer = optlib.registry[name](param, budget=60)
    optimizer.tell(ng.p.MultiobjectiveReference(), [10, 10, 10])
    for _ in range(50):
        cand = optimizer.ask()
        optimizer.tell(cand, [-38, 0, np.nan])


@pytest.mark.parametrize("name", ["LhsDE", "RandomSearch"])  # type: ignore
def test_uniform_sampling(name: str) -> None:
    param = ng.p.Scalar(lower=-100, upper=100).set_mutation(sigma=1)
    opt = optlib.registry[name](param, budget=600, num_workers=100)
    above_50 = 0
    for _ in range(100):
        above_50 += abs(opt.ask().value) > 50
    assert above_50 > 20  # should be around 50


def test_paraportfolio_de() -> None:
    workers = 40
    opt = optlib.ParaPortfolio(12, budget=100 * workers, num_workers=workers)
    for _ in range(3):
        cands = [opt.ask() for _ in range(workers)]
        for cand in cands:
            opt.tell(cand, np.random.rand())


def test_cma_logs(capsys: tp.Any) -> None:
    opt = registry["CMA"](2, budget=300, num_workers=4)
    [opt.ask() for _ in range(4)]  # pylint: disable=expression-not-assigned
    captured = capsys.readouterr()
    assert captured.out == ""
    assert captured.err == ""


def _simple_multiobjective(x):
    return [np.sum(x**2), np.sum((x - 1) ** 2)]


def test_pymoo_pf() -> None:
    optimizer = ng.optimizers.PymooNSGA2(parametrization=2, budget=300)
    optimizer.parametrization.random_state.seed(12)
    optimizer.minimize(_simple_multiobjective)
    pf = optimizer.pareto_front()
    fixed_points = [[0.25, 0.75], [0.75, 0.25]]
    for fixed_point in fixed_points:
        values = _simple_multiobjective(np.array(fixed_point))
        # check pareto front contains a candidate dominating fixed point
        assert any(
            _simple_multiobjective(x.value)[0] < values[0] and _simple_multiobjective(x.value)[1] < values[1]
            for x in pf
        )


def test_pymoo_batched() -> None:
    optimizer = ng.optimizers.PymooBatchNSGA2(parametrization=2, budget=300)
    optimizer.parametrization.random_state.seed(12)
    candidates = []
    losses = []
    optimizer.num_objectives = 2
    for _ in range(3):
        asks_from_batch = 0
        while (optimizer.num_ask == optimizer.num_tell) or asks_from_batch < optimizer.batch_size:  # type: ignore
            x = optimizer.ask()
            loss = _simple_multiobjective(*x.args, **x.kwargs)
            candidates.append(x)
            losses.append(loss)
            asks_from_batch += 1
        assert asks_from_batch == 100
        while optimizer.num_ask > optimizer.num_tell:
            x = candidates.pop()
            loss = losses.pop()
            optimizer.tell(x, loss)
    assert len(optimizer._current_batch) == 0  # type: ignore


def test_smoother() -> None:
    x = ng.p.Array(shape=(5, 5))
    assert (
        optlib.smooth_copy(x).get_standardized_data(reference=x).shape
        == x.get_standardized_data(reference=x).shape
    )
    x = ng.p.Array(shape=(5, 5)).set_integer_casting()
    assert (
        optlib.smooth_copy(x).get_standardized_data(reference=x).shape
        == x.get_standardized_data(reference=x).shape
    )


@pytest.mark.parametrize("n", [5, 10, 15, 25, 40])  # type: ignore
@pytest.mark.parametrize("b_per_dim", [10, 20])  # type: ignore
def test_voronoide(n, b_per_dim) -> None:
    if n < 25 or b_per_dim < 1 and not os.environ.get("CIRCLECI", False):  # Outside CircleCI, only the big.
        raise SkipTest("Only big things outside CircleCI.")

    list_optims = ["CMA", "DE", "PSO", "RandomSearch", "TwoPointsDE", "OnePlusOne"]
    if os.environ.get("CIRCLECI", False) and (n > 10 or n * b_per_dim > 100):  # In CircleCI, only the small.
        raise SkipTest("Topology optimization too slow in CircleCI")
    if os.environ.get("CIRCLECI", False) or (n < 10 or b_per_dim < 20):
        list_optims = ["CMA", "PSO", "OnePlusOne"]
    if n > 20:
        list_optims = ["DE", "TwoPointsDE"]
    fails = {}
    for o in list_optims:
        fails[o] = 0
    size = n * n
    sqrtsize = n
    b = b_per_dim * size  # budget
    nw = 20  # parallel workers

    num_tests = 20
    array = ng.p.Array(shape=(n, n), lower=-1.0, upper=1.0)
    for idx in range(num_tests):
        xa = idx % 3
        xb = 2 - xa
        xs = 1.5 * (
            np.array([float(np.cos(xa * i + xb * j) < 0.0) for i in range(n) for j in range(n)]).reshape(n, n)
            - 0.5
        )
        if (idx // 3) % 2 > 0:
            xs = np.transpose(xs)
        if (idx // 6) % 2 > 0:
            xs = -xs

        def f(x):
            # return np.linalg.norm(x - xs) + np.linalg.norm(x - gaussian_filter(x, sigma=1))
            return (
                5.0 * np.sum(np.abs(x - xs) > 0.3) / size
                + 13.0 * np.linalg.norm(x - gaussian_filter(x, sigma=3)) / sqrtsize
            )

        VoronoiDE = ng.optimizers.VoronoiDE(array, budget=b, num_workers=nw)
        vde = f(VoronoiDE.minimize(f).value)
        for o in list_optims:
            try:
                other = ng.optimizers.registry[o](array, budget=b, num_workers=nw)
                val = f(other.minimize(f).value)
            except:
                print(f"crash in {o}")
                val = float(1.0e7)
            # print(o, val / vde)
            if val < vde:
                fails[o] += 1
        # Remove both lines below. TODO
        # ratio = min([(idx + 1 - fails[o]) / (0.001 + fails[o]) for o in list_optims])
        # print(f"temporary: {ratio}", idx + 1, fails, f"({n}-{b_per_dim})")
    ratio = min([(num_tests - fails[o]) / (0.001 + fails[o]) for o in list_optims])
    print(f"VoronoiDE for DO: {ratio}", num_tests, fails, f"({n}-{b_per_dim})")

    for o in list_optims:
        ratio = 3.0 if "DE" not in "o" else 2.0
        assert (
            num_tests - fails[o] > ratio * fails[o]
        ), f"Failure {o}: {fails[o]} / {num_tests}    ({n}-{b_per_dim})"


def test_weighted_moo_de() -> None:
    for _ in range(1):  # Yes this is cheaper.
        D = 2
        N = 3
        DE = ng.optimizers.TwoPointsDE(D, budget=600)
        index = np.random.choice(range(N))
        w = np.ones(N)
        w[index] = 30.0
        DE.set_objective_weights(w)  # type: ignore
        targ = [np.array([np.cos(2 * np.pi * i / N), np.sin(2 * np.pi * i / N)]) for i in range(N)]
        DE.minimize(lambda x: [np.linalg.norm(x - xi) for xi in targ])  # type: ignore
        x = np.zeros(N)
        for u in DE.pareto_front():
            x = x + u.losses
        assert index == list(x).index(min(x))



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import os
import sys
import shutil
import tempfile
import subprocess
from pathlib import Path
import numpy as np
from nevergrad.common import typing as tp


class BoundChecker:
    """Simple object for checking whether an array lies
    between provided bounds.

    Parameter
    ---------
    lower: float or None
        minimum value
    upper: float or None
        maximum value

    Note
    -----
    Not all bounds are necessary (data can be partially bounded, or not at all actually)
    """

    def __init__(self, lower: tp.BoundValue = None, upper: tp.BoundValue = None) -> None:
        self.bounds = (lower, upper)

    def __call__(self, value: np.ndarray) -> bool:
        """Checks whether the array lies within the bounds

        Parameter
        ---------
        value: np.ndarray
            array to check

        Returns
        -------
        bool
            True iff the array lies within the bounds
        """
        for k, bound in enumerate(self.bounds):
            if bound is not None:
                if np.any((value > bound) if k else (value < bound)):
                    return False
        return True


class FunctionInfo:  # Note: eventually, this should be a dataclass (dropping old Python support)
    """Information about the function

    Parameters
    ----------
    deterministic: bool
        whether the function equipped with its instrumentation is deterministic.
        Can be false if the function is not deterministic or if the instrumentation
        contains a softmax.
    proxy: bool
        whether the objective function is a proxy of a more interesting objective function.
    metrizable: bool
        whether the domain is naturally equipped with a metric.
    """

    def __init__(
        self,
        deterministic: bool = True,
        proxy: bool = False,
        metrizable: bool = True,
    ) -> None:
        self.deterministic = deterministic
        self.proxy = proxy
        self.metrizable = metrizable

    def __repr__(self) -> str:
        diff = ",".join(f"{x}={y}" for x, y in sorted(self.__dict__.items()))
        return f"{self.__class__.__name__}({diff})"


_WARNING = "parameter.descriptors is deprecated use {} instead"


class TemporaryDirectoryCopy(tempfile.TemporaryDirectory):  # type: ignore
    """Creates a full copy of a directory inside a temporary directory
    This class can be used as TemporaryDirectory but:
    - the created copy path is available through the copyname attribute
    - the contextmanager returns the clean copy path
    - the directory where the temporary directory will be created
      can be controlled through the CLEAN_COPY_DIRECTORY environment
      variable
    """

    key = "CLEAN_COPY_DIRECTORY"

    @classmethod
    def set_clean_copy_environment_variable(cls, directory: tp.Union[Path, str]) -> None:
        """Sets the CLEAN_COPY_DIRECTORY environment variable in
        order for subsequent calls to use this directory as base for the
        copies.
        """
        assert Path(directory).exists(), "Directory does not exist"
        os.environ[cls.key] = str(directory)

    # pylint: disable=redefined-builtin
    def __init__(self, source: tp.Union[Path, str], dir: tp.Optional[tp.Union[Path, str]] = None) -> None:
        if dir is None:
            dir = os.environ.get(self.key, None)
        super().__init__(prefix="tmp_clean_copy_", dir=dir)
        self.copyname = Path(self.name) / Path(source).name
        shutil.copytree(str(source), str(self.copyname))

    def __enter__(self) -> Path:
        super().__enter__()
        return self.copyname


class FailedJobError(RuntimeError):
    """Job failed during processing"""


class CommandFunction:
    """Wraps a command as a function in order to make sure it goes through the
    pipeline and notify when it is finished.
    The output is a string containing everything that has been sent to stdout

    Parameters
    ----------
    command: list
        command to run, as a list
    verbose: bool
        prints the command and stdout at runtime
    cwd: Path/str
        path to the location where the command must run from

    Returns
    -------
    str
       Everything that has been sent to stdout
    """

    def __init__(
        self,
        command: tp.List[str],
        verbose: bool = False,
        cwd: tp.Optional[tp.Union[str, Path]] = None,
        env: tp.Optional[tp.Dict[str, str]] = None,
    ) -> None:
        if not isinstance(command, list):
            raise TypeError("The command must be provided as a list")
        self.command = command
        self.verbose = verbose
        self.cwd = None if cwd is None else str(cwd)
        self.env = env

    def __call__(self, *args: tp.Any, **kwargs: tp.Any) -> str:
        """Call the cammand line with addidional arguments
        The keyword arguments will be sent as --{key}={val}
        The logs are bufferized. They will be printed if the job fails, or sent as output of the function
        Errors are provided with the internal stderr
        """
        # TODO make the following command more robust (probably fails in multiple cases)
        full_command = (
            self.command + [str(x) for x in args] + ["--{}={}".format(x, y) for x, y in kwargs.items()]
        )
        if self.verbose:
            print(f"The following command is sent: {full_command}")
        outlines: tp.List[str] = []
        with subprocess.Popen(
            full_command,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            shell=False,
            cwd=self.cwd,
            env=self.env,
        ) as process:
            try:
                assert process.stdout is not None
                for line in iter(process.stdout.readline, b""):
                    if not line:
                        break
                    outlines.append(line.decode().strip())
                    if self.verbose:
                        print(outlines[-1], flush=True)
            except Exception:  # pylint: disable=broad-except
                process.kill()
                process.wait()
                raise FailedJobError("Job got killed for an unknown reason.")
            stderr = process.communicate()[1]  # we already got stdout
            stdout = "\n".join(outlines)
            retcode = process.poll()
            if stderr and (retcode or self.verbose):
                print(stderr.decode(), file=sys.stderr)
            if retcode:
                subprocess_error = subprocess.CalledProcessError(
                    retcode, process.args, output=stdout, stderr=stderr
                )
                raise FailedJobError(stderr.decode()) from subprocess_error
        return stdout


X = tp.TypeVar("X")


class Subobjects(tp.Generic[X]):
    """Identifies suboject of a class and applies
    functions recursively on them.

    Parameters
    ----------
    object: Any
        an object containing other (sub)objects
    base: Type
        the base class of the subobjects (to filter out other items)
    attribute: str
        the attribute containing the subobjects

    Note
    ----
    The current implementation is rather inefficient and could probably be
    improved a lot if this becomes critical
    """

    def __init__(self, obj: X, base: tp.Type[X], attribute: str) -> None:
        self.obj = obj
        self.cls = base
        self.attribute = attribute

    def new(self, obj: X) -> "Subobjects[X]":
        """Creates a new instance with same configuratioon
        but for a new object.
        """
        return Subobjects(obj, base=self.cls, attribute=self.attribute)

    def items(self) -> tp.Iterator[tp.Tuple[tp.Any, X]]:
        """Returns a dict {key: subobject}"""
        container = getattr(self.obj, self.attribute)
        if not isinstance(container, (list, dict)):
            raise TypeError("Subcaller only work on list and dict")
        iterator = enumerate(container) if isinstance(container, list) else container.items()
        for key, val in iterator:
            if isinstance(val, self.cls):
                yield key, val

    def _get_subobject(self, obj: X, key: tp.Any) -> tp.Any:
        """Returns the corresponding subject if obj is from the
        base class, or directly the object otherwise.
        """
        if isinstance(obj, self.cls):
            return getattr(obj, self.attribute)[key]
        return obj

    def apply(self, method: str, *args: tp.Any, **kwargs: tp.Any) -> tp.Dict[tp.Any, tp.Any]:
        """Calls the named method with the provided input parameters (or their subobjects if
        from the base class!) on the subobjects.
        """
        outputs: tp.Dict[tp.Any, tp.Any] = {}
        for key, subobj in self.items():
            subargs = [self._get_subobject(arg, key) for arg in args]
            subkwargs = {k: self._get_subobject(kwarg, key) for k, kwarg in kwargs.items()}
            outputs[key] = getattr(subobj, method)(*subargs, **subkwargs)
        return outputs


def float_penalty(x: tp.Union[bool, float]) -> float:
    """Unifies penalties as float (bool=False becomes 1).
    The value is positive for unsatisfied penality else 0.
    """
    if isinstance(x, (bool, np.bool_)):
        return float(not x)  # False ==> 1.
    elif isinstance(x, (float, np.float_)):
        return -min(0, x)  # Negative ==> >0
    raise TypeError(f"Only bools and floats are supported for check constaint, but got: {x} ({type(x)})")


class _ConstraintCompatibilityFunction:
    """temporary hack for "register_cheap_constraint", to be removed"""

    def __init__(self, func: tp.Callable[[tp.Any], tp.Loss]) -> None:
        self.func = func

    def __call__(self, *args: tp.Any, **kwargs: tp.Any) -> tp.Loss:
        out = self.func((args, kwargs))
        return out



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import typing as tp
import warnings
import numpy as np
from ..common import testing
from . import discretization


@testing.parametrized(
    arity2=(2, [0.1, -2, 0], [1, 0, 0]),
    arity100=(100, [-15, -2, -0.3, 0.3, 2, 15], [0, 2, 38, 61, 97, 99]),
    borderline_cases_100=(100, [-np.inf, np.inf, np.nan], [0, 99, 0]),
    borderline_cases_2=(2, [-np.inf, np.inf, np.nan], [0, 1, 0]),
)
def test_thresholding_discretization(arity: int, data: tp.List[float], expected: tp.List[float]) -> None:
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        output = discretization.threshold_discretization(data, arity)
    np.testing.assert_array_equal(output, expected)


def test_inverse_threshold_discretization() -> None:
    arity = 4
    indexes = np.arange(arity)  # Test all possible indexes
    data = discretization.inverse_threshold_discretization(indexes, arity)  # type: ignore
    np.testing.assert_array_equal(discretization.threshold_discretization(data, arity), indexes)


def test_encoder_probabilities() -> None:
    weights = np.array(
        [
            [0, 0, 0],
            [0, 0, 100],
            [np.nan, 0, 1],
            [np.inf, np.inf, np.inf],
            [np.inf, np.inf, 12],
            [0, -np.inf, 0],
        ]
    )
    rng = np.random.RandomState(12)
    enc = discretization.Encoder(weights, rng=rng)
    proba = enc.probabilities()
    expected = [
        [0.333, 0.333, 0.333],
        [0, 0, 1],
        [0, 0.269, 0.731],
        [0.333, 0.333, 0.333],
        [0.5, 0.5, 0],
        [0.5, 0, 0.5],
    ]
    np.testing.assert_array_almost_equal(proba, expected, decimal=3)  # type: ignore


def test_encoder() -> None:
    weights = np.array([[0, 0, 0], [0, 0, 100]])
    rng = np.random.RandomState(14)
    enc = discretization.Encoder(weights, rng=rng)
    np.testing.assert_equal(enc.encode(), [1, 2])
    np.testing.assert_equal(enc.encode(), [2, 2])
    #
    for _ in range(10):
        np.testing.assert_equal(enc.encode(True), [0, 2])
